import json

from google.cloud.firestore_v1 import Query, CollectionReference

from scoutr.exceptions import BadRequestException
from scoutr.providers.base.filtering import Filtering


class GCPFiltering(Filtering):
    def __init__(self, collection: CollectionReference):
        self.collection = collection
        self.query = Query(collection)

    def And(self, condition1: Query, condition2: Query):
        if condition1 and condition2:
            return Query(
                self.collection,
                field_filters=condition1._field_filters + condition2._field_filters
            )
        elif condition1:
            return condition1
        elif condition2:
            return condition2

    def Or(self, condition1: Query, condition2: Query):
        if condition1 and condition2:
            return condition1 | condition2
        elif condition1:
            return condition1
        elif condition2:
            return condition2
        else:
            return None

    def equals(self, attr: str, value) -> Query:
        return Query(self.collection).where(attr, '==', value)

    def not_equal(self, attr: str, value) -> Query:
        return Query(self.collection).where(attr, '!=', value)

    def greater_than(self, attr: str, value) -> Query:
        return Query(self.collection).where(attr, '>', value)

    def less_than(self, attr: str, value) -> Query:
        return Query(self.collection).where(attr, '<', value)

    def greater_than_equal(self, attr: str, value) -> Query:
        return Query(self.collection).where(attr, '>=', value)

    def less_than_equal(self, attr: str, value) -> Query:
        return Query(self.collection).where(attr, '<=', value)

    def between(self, attr: str, value) -> Query:
        if isinstance(value, list):
            values = value
        else:
            values = json.loads(value)
        if not len(values) == 2:
            raise BadRequestException('Between operation requires two values')

        return Query(self.collection).where(attr, '>=', values[0]).where(attr, '<=', values[1])

    def is_in(self, attr: str, value) -> Query:
        if isinstance(value, list):
            values = value
        else:
            values = json.loads(value)

        return Query(self.collection).where(attr, 'in', values)

    def not_in(self, attr: str, value) -> Query:
        if isinstance(value, list):
            values = value
        else:
            values = json.loads(value)

        return Query(self.collection).where(attr, 'not-in', values)



from typing import get_type_hints, Dict, Any, _GenericAlias


class Model:
    def __init__(self, **kwargs):
        for attr, cls in get_type_hints(self.__class__).items():
            # Convert types
            try:
                if isinstance(cls, _GenericAlias):
                    if cls._name == 'List':
                        cls = list
                    elif cls._name == 'Dict':
                        cls = dict
            except TypeError:
                pass

            # Check if attribute is not in kwargs
            if attr not in kwargs:
                # If a default value is set, use that
                args = []
                default_val = getattr(self, attr, None)
                if default_val is not None:
                    args.append(default_val)

                # Try to initialize the value
                try:
                    value = cls(*args)
                except TypeError:
                    value = None
            else:
                # Set value from kwargs
                value = kwargs[attr]

                # Detect if this is a Model class
                try:
                    is_model_class = issubclass(cls, Model) and isinstance(value, dict)
                except TypeError:
                    is_model_class = False

                # Run Modal class constructor
                if is_model_class:
                    value = cls(**value)

            # Set the attribute
            setattr(self, attr, value)

    @classmethod
    def attributes(cls) -> tuple:
        # attributes = inspect.getmembers(cls, lambda a: not (inspect.isroutine(a)))
        # return [a[0] for a in attributes if not (a[0].startswith('__') and a[0].endswith('__'))]
        return tuple(get_type_hints(cls).keys())

    def dict(self) -> Dict[str, Any]:
        output = {}
        for attribute in self.attributes():
            value = getattr(self, attribute)
            if isinstance(value, list):
                for i, item in enumerate(value):
                    if isinstance(item, Model):
                        value[i] = item.dict()
            elif isinstance(value, Model):
                value = value.dict()

            if value is not None:
                output[attribute] = value

        return output

    @classmethod
    def load(cls, data: Dict[str, Any], skip_validation=False):
        model = cls(**data)
        if not skip_validation:
            model.validate()

        return model

    def validate(self):
        valid_attributes = set(self.attributes())

        # Compile list of required fields (fields that default to None)
        required_fields = set()
        for attr in valid_attributes:
            if getattr(self.__class__, attr, None) is None:
                required_fields.add(attr)

        # Compile list of fields that have values
        missing_fields = set()
        for attr in valid_attributes:
            if getattr(self, attr, None) is None:
                missing_fields.add(attr)

        # Make sure required fields have values
        if missing_fields:
            raise Exception(f'Missing required fields on {self.__class__.__name__}: {missing_fields}')



from typing import List, Any, Optional

from scoutr.models import Model


class UserData(Model):
    username: str
    name: str
    email: str
    groups: List[str]


class RequestUser(Model):
    id: str
    data: Optional[UserData] = None


class Request(Model):
    user: RequestUser
    method: str
    path: str
    source_ip: str
    user_agent: str
    body: Any
    path_params: dict
    query_params: dict



import re

from scoutr.providers.aws import DynamoAPI


"""
Valid option groups
"""
VALID_STATUSES = [
    'Active',
    'Inactive'
]

"""
Fields required for record creation
"""
CREATE_FIELDS = {
    'id': lambda value, item, existing_item: {
        'result': re.match(r'^\d{6}', value),
        'message': 'Invalid id'
    },
    'description': lambda value, item, existing_item: isinstance(value, str),
    'status': lambda value, item, existing_item: DynamoAPI.value_in_set(
        value=value,
        valid_options=VALID_STATUSES,
        option_name='status'
    )
}

"""
Field validation for record updates
"""
UPDATE_FIELDS = {
    'description': lambda value, item, existing_item: isinstance(value, str),
    'type': lambda value, item, existing_item: {
        'result': existing_item.get('type') == item.get('type') if 'type' in item else False,
        'message': 'Type cannot be modified'
    }
}

"""
Fields that cannot be updated
"""
RESTRICTED_FIELDS = [
    'id',
    'type'
]
UPDATE_FIELDS.update({
    field: lambda value, item, existing_item, field_name=field: {
        'result': False,
        'message': f'{field_name} cannot be updated'
    }
    for field in RESTRICTED_FIELDS
})



import logging
from typing import List, Optional, Any, Callable, Union, Tuple

from bson import ObjectId
from pymongo import MongoClient
from pymongo.errors import DuplicateKeyError

from scoutr.exceptions import NotFoundException, BadRequestException, UnauthorizedException
from scoutr.models.config import MongoConfig
from scoutr.models.request import Request
from scoutr.models.user import User, Group
from scoutr.providers.base import BaseAPI
from scoutr.providers.mongo.filtering import MongoFiltering

try:
    import sentry_sdk
except ImportError:
    from scoutr.utils import mock_sentry
    sentry_sdk = mock_sentry

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class MongoAPI(BaseAPI):
    filtering = MongoFiltering()

    def __init__(self, config: MongoConfig):
        super(MongoAPI, self).__init__(config)
        self.config = config
        self.client = MongoClient(config.connection_string)
        self.db = self.client.get_database(self.config.database)
        self.data_table = self.db.get_collection(self.config.data_table)
        self.auth_table = self.db.get_collection(self.config.auth_table)
        self.group_table = self.db.get_collection(self.config.group_table)
        self.audit_table = self.db.get_collection(self.config.audit_table)

    def get_auth(self, user_id: str) -> Optional[User]:
        # Try to find user in the auth table
        result = self.auth_table.find_one(user_id)

        if not result:
            return None

        # Create user object
        return User.load(result)

    def get_entitlements(self, entitlement_ids: List[str]) -> List[User]:
        entitlements = []
        conditions = self.filtering.is_in('id', entitlement_ids)

        for record in self.data_table.find(conditions):
            # Convert ObjectId to string if necessary
            if isinstance(record['_id'], ObjectId):
                record['_id'] = str(record['_id'])

            # Overwrite the id
            record['id'] = record['_id']

            # Add the record
            entitlements.append(User.load(record))

        return entitlements

    def get_group(self, group_id: str) -> Optional[Group]:
        # Try to find user in the auth table
        result = self.group_table.find_one(group_id)

        if not result:
            return None

        # Create user object
        return Group.load(result)

    def store_item(self, table: str, item: dict) -> bool:
        try:
            self.db.get_collection(table).insert_one(item)
        except Exception as e:
            print('Failed to store record: %s', e)
            return False

        return True

    @staticmethod
    def get_projection_expression(user: User) -> dict:
        # Build a projection expression to exclude fields based on user permissions
        projection_expression = {}
        for excluded_field in user.exclude_fields:
            projection_expression[excluded_field] = False

        # If any fields are being excluded, set the projection expression
        args = {}
        if projection_expression:
            args['projection'] = projection_expression

        return args

    def create(self, request: Request, data: dict, validation: dict = None,
               required_fields: Union[List, Tuple] = ()) -> dict:
        """
        Create an item

        :param Request request: Request object
        :param dict data: Item to create
        :param dict validation: Optional dictionary containing mappings of field name to callable. See the docstring
        in the _validate_fields method for more information.
        :param list required_fields: List of required fields
        :return: Created item
        :rtype: dict
        """
        user = self._prepare_create(request, data, validation, required_fields)

        # Make sure primary key is included
        document_id = data.get(self.config.primary_key)
        if not document_id:
            raise BadRequestException('Primary key %s is required' % self.config.primary_key)

        # Set the resource identifier
        resource = {self.config.primary_key: document_id}

        # TODO: Build condition to ensure the unique key does not exist

        # Save the item
        try:
            # Set the document id
            data['_id'] = document_id

            # Insert the record
            result = self.data_table.insert_one(data)
            if not result.inserted_id:
                raise Exception('Failed to save item')

            logger.info('[%(user)s] Successfully created item:\n%(item)s' % {
                'user': self.user_identifier(user),
                'item': data
            })
        except DuplicateKeyError:
            raise BadRequestException(f"An item with id '{document_id}' already exists")

        except Exception as e:
            logger.error(
                '[%(user)s] Encountered error while attempting to create record %(error)s. Item:\n%(item)s' % {
                    'user': self.user_identifier(user),
                    'error': str(e),
                    'item': data
                }
            )
            raise

        sentry_sdk.add_breadcrumb(category='data', message='Created item in Firestore', level='info')

        # Create audit log
        self.audit_log(action=self.AUDIT_ACTION_CREATE, resource=resource, request=request, user=user)

        return resource

    def update(self, request: Request, primary_key: dict, data: dict, validation: dict,
               audit_action=BaseAPI.AUDIT_ACTION_UPDATE) -> dict:
        """
        Update an item

        :param Request request: Request object
        :param dict primary_key: Dictionary formatted as {"primary_key": "value_of_row_to_update"}
        :param dict data: Fields to update, formatted as {"key": "value"}
        :param dict validation: Optional dictionary containing mappings of field name to callable. See the
        docstring in the _validate_fields method for more information.
        :param str audit_action: Action to use in the audit log. This defaults to UPDATE, and is provided as a
        convenience to the user in case customizing the phrasing is desired. This cannot be one of the reserved
        built-in actions: CREATE, DELETE, GET, LIST, SEARCH.
        :return: Updated item
        :rtype: dict
        """
        # Get user
        user = self.initialize_request(request)

        # Validate audit action
        audit_action = audit_action.upper()
        if audit_action in (self.AUDIT_ACTION_CREATE, self.AUDIT_ACTION_DELETE,
                            self.AUDIT_ACTION_GET, self.AUDIT_ACTION_LIST, self.AUDIT_ACTION_SEARCH):
            raise Exception('%s is a reserved built-in audit action' % audit_action)

        # Deny updates to primary key
        if self.config.primary_key in data:
            raise BadRequestException('Primary key %s cannot be updated' % self.config.primary_key)

        # Add in the user's permissions
        conditions = self.filtering.filter(
            user,
            {'_id': primary_key[self.config.primary_key]},
            action=self.filtering.FILTER_ACTION_READ
        )

        # Get the existing item
        existing_item = self.data_table.find_one(conditions)
        if len(existing_item) == 0:
            logger.info('[%(user)s] Primary key "%(primary_key)s" does not exist or user does '
                        'not have permission to access it' % {
                            'user': self.user_identifier(user),
                            'primary_key': primary_key
                        })
            raise NotFoundException('Item does not exist or you do not have permission to access it')
        elif len(existing_item) > 1:
            # Should not have found more than one item
            raise BadRequestException('Multiple items found')

        # Found the item
        existing_item = existing_item[0]

        # Validate update
        self.validate_update(user, data, existing_item)

        # Perform field validation
        if validation:
            self.validate_fields(
                validation=validation,
                required_fields=(),
                item=data,
                existing_item=existing_item
            )
            sentry_sdk.add_breadcrumb(category='validate', message='Validated input fields', level='info')

        # Perform the update item call
        try:
            document_id = primary_key[self.config.primary_key]
            self.data_table.update_one(conditions, {'$set': data})

            # Build a projection expression to exclude fields based on user permissions
            args = self.get_projection_expression(user)

            # Pull the updated item from the DB
            response = self.data_table.find_one(document_id, **args)
            logger.info('[%(user)s] Successfully updated record "%(primary_key)s" with values:\n%(item)s' % {
                'user': self.user_identifier(user),
                'primary_key': primary_key,
                'item': data
            })
        except Exception as e:
            logger.error(
                '[%(user)s] Encountered error while attempting to update record '
                '"%(primary_key)s": %(error)s. Item:\n%(item)s' % {
                    'user': self.user_identifier(user),
                    'primary_key': primary_key,
                    'error': str(e),
                    'item': data
                }
            )
            raise

        sentry_sdk.add_breadcrumb(category='data', message='Updated item', level='info')

        # Create audit log
        self.audit_log(action=audit_action, resource=primary_key, changes=data, request=request, user=user)

        # Return updated record
        return response

    def get(self, request: Request, key: Any, value: Any) -> dict:
        """
        Get a specific item from the table

        :param Request request: Request object
        :param str key: Key to search on
        :param str value: Value to search for
        :return: Item
        :rtype: dict
        """
        # Get user
        user = self.initialize_request(request)

        # Filter the data according to the user's permissions
        conditions = self.filtering.And(
            self.filtering.filter(user),
            self.filtering.equals(key, value)
        )

        # Search for the item
        doc_count = self.data_table.count_documents(conditions)

        # Check results
        if doc_count == 0:
            raise NotFoundException('Item not found')
        elif doc_count > 1:
            # This should only be returning a single item
            raise BadRequestException('Multiple items returned')

        # Build a projection expression to exclude fields based on user permissions
        args = self.get_projection_expression(user)

        # Run the query
        data = self.data_table.find_one(conditions, **args)
        sentry_sdk.add_breadcrumb(
            category='query', message='%s = %s' % (key, value), level='info', table=self.config.data_table
        )

        # Convert ObjectId to string if necessary
        if isinstance(data['_id'], ObjectId):
            data['_id'] = str(data['_id'])

        # Item was found, return the single item
        self.audit_log(action=self.AUDIT_ACTION_GET, request=request, user=user, resource={key: value})
        return data

    def list(self, request: Request) -> List[dict]:
        """
        List all values in a table

        :param Request request: Request object
        :return: Data from the table
        :rtype: dict
        """
        user, params = self._prepare_list(request)

        # Build filters
        conditions = self.filtering.filter(user, params)
        if conditions is None:
            conditions = {}

        # Build a projection expression to exclude fields based on user permissions
        args = self.get_projection_expression(user)

        # Perform the query
        data = []
        for record in self.data_table.find(conditions, **args):
            # Convert ObjectId to string if necessary
            if isinstance(record['_id'], ObjectId):
                record['_id'] = str(record['_id'])

            # Add the record
            data.append(record)

        # Add audit log
        self.audit_log(self.AUDIT_ACTION_LIST, request, user)

        return data

    def list_unique_values(self, request: Request, key: str,
                           unique_func: Callable[[List, str], List[str]] = lambda items, _: sorted(items)) -> List[str]:
        user, params = self._prepare_list(request, False)

        # Build filters
        conditions = self.filtering.And(
            self.filtering.filter(user, params),
            self.filtering.exists(key, 'true')
        )

        # Build a projection expression to exclude fields based on user permissions
        args = self.get_projection_expression(user)

        # Download the data
        try:
            raw_data = self.data_table.find(conditions, **args)
            distinct_values = raw_data.distinct(key)
        except Exception as e:
            logger.error(
                '[%(user)s] Encountered error while attempting to list records: %(error)s' % {
                    'user': self.user_identifier(user),
                    'error': str(e)
                }
            )
            raise

        # Make sure the requested key exists in the output - we only need to check the first record in the cursor.
        # It will be missing if user does not have permissions to view it
        for item in raw_data:
            if key not in item:
                raise UnauthorizedException(f"Not authorized to view contents of field '{key}'")
            break

        # Use the distinct values that were already saved
        data = distinct_values

        # Make sure a unique, sorted list is returned
        output = unique_func(data, key)

        # Create audit log
        self.audit_log(self.AUDIT_ACTION_LIST, request, user)

        return output

    def list_audit_logs(self, request: Request, param_overrides: dict = None) -> List[dict]:
        """
        List all audit logs

        :param Request request: Request object
        :param dict param_overrides: Parameter overrides: Each parameter should be formatted as:

            {
                "field_name": "search_value"
            }

        :return: List of audit logs, sorted with most recent entry first
        :rtype: list of dict
        """
        if not self.config.audit_table:
            raise NotFoundException('Audit logs are not enabled')

        if param_overrides is None:
            param_overrides = {}

        # Merge parameter overrides into query params
        request.query_params.update(param_overrides)

        # Get user
        user = self.initialize_request(request)

        # Build filters
        conditions = self.filtering.filter(None, request.query_params)

        # Download data
        try:
            data = self.audit_table.find(conditions)
        except Exception as e:
            logger.error(
                '[%(user)s] Encountered error while attempting to list audit logs: %(error)s' % {
                    'user': self.user_identifier(user),
                    'error': str(e)
                }
            )
            raise

        # Add sentry breadcrumb
        sentry_sdk.add_breadcrumb(
            category='query',
            message='Scanned table',
            level='info',
            table=self.config.audit_table
        )

        # Sort the data
        data = sorted(data, key=lambda item: item['time'], reverse=True)

        return data

    def search(self, request: Request, key: str, values: List[str]) -> List[dict]:
        """
        Perform a multi-value search of a field in the table. The search endpoint should be configured in API Gateway:

            POST /search/{search_key}

        The value of {search_key} should be passed in as `key` and the contents of the POST request body should be
        passed in as the `values` list.

        :param Request request: Request object
        :param str key: Field to search
        :param list of str values: Search values
        :return: Search results
        :rtype: dict
        """
        # Get user
        user = self.initialize_request(request)

        # Build multi-value filter expressions
        multi_filters = self.filtering.multi_filter(user, key, values)
        sentry_sdk.add_breadcrumb(category='data', message='Built multi-value filter', level='info')

        # Build a projection expression to exclude fields based on user permissions
        args = self.get_projection_expression(user)

        # Perform each generated filter and then combine the results together
        output = []
        for filters in multi_filters:
            results = self.data_table.find(filters, **args)
            if not results:
                continue

            if isinstance(results, dict):
                results = [results]

            for record in results:
                # Convert ObjectId to string if necessary
                if isinstance(record['_id'], ObjectId):
                    record['_id'] = str(record['_id'])
                output.append(record)

        # Create audit log
        self.audit_log(action=self.AUDIT_ACTION_SEARCH, request=request, user=user)

        # Return the results
        return output

    def delete(self, request: Request, primary_key: dict) -> dict:
        """
        Delete an item

        :param Request request: Request object
        :param dict primary_key: Dictionary formatted as {"primary_key": "value_of_row_to_delete"}
        :return: Success
        :rtype: dict
        """
        # Get user
        user = self.initialize_request(request)

        # Add in the user's permissions
        conditions = self.filtering.And(
            self.filtering.filter(user, action=self.filtering.FILTER_ACTION_DELETE),
            self.filtering.equals('_id', primary_key[self.config.primary_key])
        )

        # Perform the deletion
        try:
            result = self.data_table.delete_one(conditions)
            if result.deleted_count == 0:
                raise NotFoundException('Failed to delete item')

            logger.info('[%(user)s] Successfully deleted record "%(primary_key)s"' % {
                'user': self.user_identifier(user),
                'primary_key': primary_key,
            })
        except Exception as e:
            logger.error(
                '[%(user)s] Encountered error while attempting to delete record "%(primary_key)s": %(error)s' % {
                    'user': self.user_identifier(user),
                    'primary_key': primary_key,
                    'error': str(e)
                }
            )
            raise

        sentry_sdk.add_breadcrumb(category='data', message='Deleted item', level='info')

        # Create audit log
        self.audit_log(
            action=self.AUDIT_ACTION_DELETE,
            request=request,
            user=user,
            resource=primary_key,
        )

        # Return updated record
        return {'deleted': True}



import simplejson as json
import os

import sentry_sdk

from example.utils import get_config
from scoutr.helpers.api_gateway import build_api_gateway_request
from scoutr.providers.aws import DynamoAPI
from scoutr.exceptions import HttpException
from example.utils import configure_sentry

configure_sentry()


def main(event, context):
    # Build request
    request = build_api_gateway_request(event)

    try:
        api = DynamoAPI(get_config())

        # Validate item id
        item = request.path_params.get('item')
        if item:
            param_overrides = {'resource.id': item}
        else:
            param_overrides = {}

        # Perform the search
        data = api.list_audit_logs(request=request, param_overrides=param_overrides)

    except HttpException as e:
        if e.status == 500:
            sentry_sdk.capture_exception(e)
        return {
            'statusCode': e.status,
            'body': json.dumps({'error': str(e)})
        }

    except Exception as e:
        sentry_sdk.capture_exception(e)
        return {
            'statusCode': 500,
            'body': json.dumps({'error': '%s: %s' % (e.__class__.__name__, str(e))})
        }

    return {
        'statusCode': 200,
        'body': json.dumps(data)
    }



import logging
from typing import List, Dict, Optional, Any, Callable, Tuple, Union

import boto3
from mypy_boto3_dynamodb.service_resource import DynamoDBServiceResource
from boto3.dynamodb.conditions import Attr
from botocore.exceptions import ClientError

from scoutr.exceptions import NotFoundException, BadRequestException
from scoutr.models.config import Config
from scoutr.models.request import Request
from scoutr.models.user import User, Group
from scoutr.providers.aws.filtering import AWSFiltering
from scoutr.providers.base.api import BaseAPI

try:
    import sentry_sdk
except ImportError:
    from scoutr.utils import mock_sentry

    sentry_sdk = mock_sentry

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class DynamoAPI(BaseAPI):
    filtering = AWSFiltering()

    def __init__(self, config: Config):
        super(DynamoAPI, self).__init__(config)
        resource: DynamoDBServiceResource = boto3.resource('dynamodb')
        self.resource = resource
        self.data_table = resource.Table(self.config.data_table)
        self.auth_table = resource.Table(self.config.auth_table)
        self.group_table = resource.Table(self.config.group_table)
        self.audit_table = resource.Table(self.config.audit_table)

    def get_auth(self, user_id: str) -> Optional[User]:
        # Try to find user in the auth table
        result = self.auth_table.get_item(Key={'id': user_id})

        if not result.get('Item'):
            return None

        # Create user object
        return User.load(result['Item'])

    def get_entitlements(self, entitlement_ids: List[str]) -> List[User]:
        start_index = 0
        conditions = None

        # IN expressions are limited to 100 items each
        for end_index in range(0, len(entitlement_ids), 100):
            # Create a slice of 100 items
            items = entitlement_ids[start_index:end_index]

            # Skip if no items are in this slice
            if not items:
                continue

            # Create IN expression
            expr = Attr('id').is_in(items)

            # Combine with conditions using OR
            if conditions:
                conditions |= expr
            else:
                conditions = expr

            # Set new start index
            start_index = end_index

        # Add any extra items at the end
        if len(entitlement_ids[end_index:]) > 0:
            conditions |= Attr('id').is_in(entitlement_ids[end_index:])

        # Scan for the entitlement ids
        results = self._scan(self.auth_table, FilterExpression=conditions)

        # Convert to User instances
        entitlements: List[User] = []
        for item in results:
            auth = User.load(item, skip_validation=True)
            entitlements.append(auth)

        return entitlements

    def get_group(self, group_id: str) -> Optional[Group]:
        # Try to find user in the auth table
        result = self.group_table.get_item(Key={'id': group_id})

        if not result.get('Item'):
            return None

        # Create user object
        return Group.load(result['Item'])

    def store_item(self, table: str, item: dict) -> bool:
        try:
            self.resource.Table(table).put_item(Item=item)
        except ClientError as e:
            print('Failed to store record: %s', e)
            return False

        return True

    @staticmethod
    def _scan(resource, **kwargs):
        response = resource.scan(**kwargs)
        items = response['Items']
        while response.get('LastEvaluatedKey', False):
            kwargs.update({'ExclusiveStartKey': response['LastEvaluatedKey']})
            response = resource.scan(**kwargs)
            items.extend(response['Items'])
        return items

    def create(self, request: Request, data: dict, validation: dict = None,
               required_fields: Union[List, Tuple] = (), condition=None, condition_failure_message='') -> dict:
        """
        Create an item

        :param Request request: Request object
        :param dict data: Item to create
        :param dict validation: Optional dictionary containing mappings of field name to callable. See the docstring
        in the _validate_fields method for more information.
        :param list required_fields: Optional list of required fields
        :param condition: Optional condition expression to apply to PutItem call
        :param str condition_failure_message: Optional condition failure message to use instead of the default
        :return: Created item
        :rtype: dict
        """
        user = self._prepare_create(request, data, validation, required_fields)

        # Build condition to ensure the unique key does not exist
        resource: Dict[str, str] = {}
        conditions = None  # self.filtering.filter(user, action=self.filtering.FILTER_ACTION_CREATE)
        for schema in self.data_table.key_schema:
            resource.update({schema['AttributeName']: data.get(schema['AttributeName'])})
            conditions = self.filtering.And(
                conditions,
                Attr(schema['AttributeName']).not_exists()
            )

        # Apply user's condition if it exists
        if condition:
            conditions = self.filtering.And(
                conditions,
                condition
            )

        try:
            self.data_table.put_item(Item=data, ConditionExpression=conditions)
            logger.info('[%(user)s] Successfully created item:\n%(item)s' % {
                'user': self.user_identifier(user),
                'item': data
            })
        except ClientError as e:
            if e.response['Error']['Code'] == 'ConditionalCheckFailedException':
                logger.warning('[%(user)s] Unable to create item because the conditional check failed:\n%(item)s' % {
                    'user': self.user_identifier(user),
                    'item': data
                })
                if condition_failure_message:
                    raise BadRequestException(condition_failure_message)
                else:
                    raise BadRequestException('Item already exists or you do not have permission to create it.')
            elif e.response['Error']['Code'] == 'ValidationException':
                logger.error('[%(user)s] Validation error - %(error)s:\n%(item)s' % {
                    'user': self.user_identifier(user),
                    'error': e.response['Error']['Message'],
                    'item': data
                })
                raise BadRequestException(e.response['Error']['Message'])
            logger.error(
                '[%(user)s] Encountered error while attempting to create record '
                '[%(code)s] %(error)s. Item:\n%(item)s' % {
                    'user': self.user_identifier(user),
                    'code': e.response['Error']['Code'],
                    'error': e.response['Error']['Message'],
                    'item': data
                }
            )
            raise
        except Exception as e:
            logger.error(
                '[%(user)s] Encountered error while attempting to create record %(error)s. Item:\n%(item)s' % {
                    'user': self.user_identifier(user),
                    'error': str(e),
                    'item': data
                }
            )
            raise

        sentry_sdk.add_breadcrumb(category='data', message='Created item', level='info')

        # Create audit log
        self.audit_log(action=self.AUDIT_ACTION_CREATE, resource=resource, request=request, user=user)

        return resource

    def update(self, request: Request, primary_key: dict, data: dict, validation: dict = None, condition=None,
               condition_failure_message='', audit_action=BaseAPI.AUDIT_ACTION_UPDATE) -> dict:
        """
        Update an item

        :param Request request: Request object
        :param dict primary_key: Dictionary formatted as {"primary_key": "value_of_row_to_update"}
        :param dict data: Fields to update, formatted as {"key": "value"}
        :param dict validation: Optional dictionary containing mappings of field name to callable. See the
        docstring in the _validate_fields method for more information.
        :param dict condition: Optional condition expression to apply to this update. If the condition fails to return
        successful, then this item will not be updated.
        :param str condition_failure_message: If the conditional check fails, this optional error message
        will be displayed to the user.
        :param str audit_action: Action to use in the audit log. This defaults to UPDATE, and is provided as a
        convenience to the user in case customizing the phrasing is desired. This cannot be one of the reserved
        built-in actions: CREATE, DELETE, GET, LIST, SEARCH.
        :return: Updated item
        :rtype: dict
        """
        # Get user
        user = self.initialize_request(request)

        # Validate audit action
        audit_action = audit_action.upper()
        if audit_action in (self.AUDIT_ACTION_CREATE, self.AUDIT_ACTION_DELETE,
                            self.AUDIT_ACTION_GET, self.AUDIT_ACTION_LIST, self.AUDIT_ACTION_SEARCH):
            raise Exception('%s is a reserved built-in audit action' % audit_action)

        # Add in the user's permissions
        conditions = self.filtering.filter(user, action=self.filtering.FILTER_ACTION_UPDATE)

        # Get the existing item / make sure it actually exists and user has permission to access it
        for key, value in primary_key.items():
            # Check if the primary key is specified in the data input
            if key in data:
                raise BadRequestException('Primary key cannot be updated')

            conditions = self.filtering.And(conditions, Attr(key).eq(value))

        # Get the existing item
        existing_item = self._scan(self.data_table, FilterExpression=conditions)
        if len(existing_item) == 0:
            logger.info('[%(user)s] Primary key "%(primary_key)s" does not exist or user does '
                        'not have permission to access it' % {
                            'user': self.user_identifier(user),
                            'primary_key': primary_key
                        })
            raise NotFoundException('Item does not exist or you do not have permission to access it')
        elif len(existing_item) > 1:
            # Should not have found more than one item
            raise BadRequestException('Multiple items found')

        # Found the item
        existing_item = existing_item[0]

        # Check user update permissions
        self.validate_update(user, data, existing_item)

        # Perform field validation
        if validation:
            self.validate_fields(
                validation=validation,
                required_fields=(),
                item=data,
                existing_item=existing_item
            )
            sentry_sdk.add_breadcrumb(category='validate', message='Validated input fields', level='info')

        # Build the update expression
        updates = []
        names = {}
        values = {}
        for key, value in data.items():
            names['#' + key] = key
            values[':' + key] = value
            updates.append(f'#{key} = :{key}')
        update_expression = 'SET ' + ', '.join(updates)

        # Perform the update item call
        try:
            args = {}
            if condition:
                args['ConditionExpression'] = condition

            response = self.data_table.update_item(
                Key=primary_key,
                UpdateExpression=update_expression,
                ExpressionAttributeNames=names,
                ExpressionAttributeValues=values,
                ReturnValues='ALL_NEW',
                **args
            )
            logger.info('[%(user)s] Successfully updated record "%(primary_key)s" with values:\n%(item)s' % {
                'user': self.user_identifier(user),
                'primary_key': primary_key,
                'item': data
            })
        except ClientError as e:
            if e.response['Error']['Code'] == 'ConditionalCheckFailedException':
                logger.warning('[%(user)s] Unable to update item because the conditional check failed:\n%(item)s' % {
                    'user': self.user_identifier(user),
                    'item': data
                })
                raise BadRequestException(condition_failure_message or 'Conditional check failed')
            logger.error(
                '[%(user)s] Encountered error while attempting to update record '
                '"%(primary_key)s": [%(code)s] %(error)s. Item:\n%(item)s' % {
                    'user': self.user_identifier(user),
                    'primary_key': primary_key,
                    'code': e.response['Error']['Code'],
                    'error': e.response['Error']['Message'],
                    'item': data
                }
            )
            raise
        except Exception as e:
            logger.error(
                '[%(user)s] Encountered error while attempting to update record '
                '"%(primary_key)s": %(error)s. Item:\n%(item)s' % {
                    'user': self.user_identifier(user),
                    'primary_key': primary_key,
                    'error': str(e),
                    'item': data
                }
            )
            raise

        sentry_sdk.add_breadcrumb(category='data', message='Updated item', level='info')

        # Create audit log
        self.audit_log(action=audit_action, resource=primary_key, changes=data, request=request, user=user)

        # Return updated record
        return self.post_process([response['Attributes']], user)[0]

    def get(self, request: Request, key: Any, value: Any) -> dict:
        """
        Get a specific item from the table

        :param Request request: Request object
        :param str key: Key to search on
        :param str value: Value to search for
        :return: Item
        :rtype: dict
        """
        # Get user
        user = self.initialize_request(request)

        # Filter the data according to the user's permissions
        conditions = self.filtering.And(
            self.filtering.filter(user),
            self.filtering.equals(key, value)
        )

        # Search for the item
        data = self._scan(self.data_table, FilterExpression=conditions)
        sentry_sdk.add_breadcrumb(
            category='query', message='%s = %s' % (key, value), level='info', table=self.config.data_table
        )

        # There should only be a single item returned
        if len(data) == 0:
            raise NotFoundException('Item not found')
        elif len(data) > 1:
            # This should only be returning a single item
            raise BadRequestException('Multiple items returned')

        # Filter the response
        output = self.post_process(data, user)[0]

        # Item was found, return the single item
        self.audit_log(action=self.AUDIT_ACTION_GET, request=request, user=user, resource={key: value})
        return output

    def list(self, request: Request) -> List[dict]:
        """
        List all values in a table

        :param Request request: Request object
        :return: Data from the table
        :rtype: dict
        """
        user, params = self._prepare_list(request)

        # Build filters
        args = {}
        conditions = self.filtering.filter(user, params)
        if conditions:
            args.update({'FilterExpression': conditions})

        try:
            # Scan the table
            data = self._scan(self.data_table, **args)
        except ClientError as e:
            logger.error(
                '[%(user)s] Encountered error while attempting to list records: [%(code)s] %(error)s' % {
                    'user': self.user_identifier(user),
                    'code': e.response['Error']['Code'],
                    'error': e.response['Error']['Message']
                }
            )
            raise
        except Exception as e:
            logger.error(
                '[%(user)s] Encountered error while attempting to list records: %(error)s' % {
                    'user': self.user_identifier(user),
                    'error': str(e)
                }
            )
            raise

        # Perform post processing
        data = self.post_process(data, user)

        # Add audit log
        self.audit_log(self.AUDIT_ACTION_LIST, request, user)

        return data

    def list_unique_values(self, request: Request, key: str,
                           unique_func: Callable[[List, str], List[str]] = BaseAPI.unique_func) -> List[str]:
        user, params = self._prepare_list(request, False, False)

        # Build filters
        conditions = self.filtering.And(
            self.filtering.filter(user, params),
            self.filtering.exists(key, 'true')
        )

        # Download the data
        try:
            data = self._scan(self.data_table, FilterExpression=conditions)
        except ClientError as e:
            logger.error(
                '[%(user)s] Encountered error while attempting to list records: [%(code)s] %(error)s' % {
                    'user': self.user_identifier(user),
                    'code': e.response['Error']['Code'],
                    'error': e.response['Error']['Message']
                }
            )
            raise
        except Exception as e:
            logger.error(
                '[%(user)s] Encountered error while attempting to list records: %(error)s' % {
                    'user': self.user_identifier(user),
                    'error': str(e)
                }
            )
            raise

        # Post process the data
        data = self.post_process(data, user)

        # Make sure a unique, sorted list is returned
        output = unique_func(data, key)

        # Create audit log
        self.audit_log(self.AUDIT_ACTION_LIST, request, user)

        return output

    def list_audit_logs(self, request: Request, param_overrides: dict = None) -> List[dict]:
        """
        List all audit logs

        :param Request request: Request object
        :param dict param_overrides: Parameter overrides: Each parameter should be formatted as:

            {
                "field_name": "search_value"
            }

        :return: List of audit logs, sorted with most recent entry first
        :rtype: list of dict
        """
        if not self.config.audit_table:
            raise NotFoundException('Audit logs are not enabled')

        if param_overrides is None:
            param_overrides = {}

        # Merge parameter overrides into query params
        request.query_params.update(param_overrides)

        # Get user
        user = self.initialize_request(request)

        # Build filters
        args = {}
        conditions = self.filtering.filter(None, request.query_params)
        if conditions:
            args['FilterExpression'] = conditions

        # Download data
        try:
            data = self._scan(self.audit_table, **args)
        except ClientError as e:
            logger.error(
                '[%(user)s] Encountered error while attempting to list audit logs: [%(code)s] %(error)s' % {
                    'user': self.user_identifier(user),
                    'code': e.response['Error']['Code'],
                    'error': e.response['Error']['Message']
                }
            )
            raise
        except Exception as e:
            logger.error(
                '[%(user)s] Encountered error while attempting to list audit logs: %(error)s' % {
                    'user': self.user_identifier(user),
                    'error': str(e)
                }
            )
            raise

        # Add sentry breadcrumb
        sentry_sdk.add_breadcrumb(
            category='query',
            message=args.get('FilterExpression', 'Scanned table'),
            level='info',
            table=self.config.audit_table
        )

        # Sort the data
        data = sorted(data, key=lambda item: item['time'], reverse=True)

        return data

    def search(self, request: Request, key: str, values: List[str]) -> List[dict]:
        """
        Perform a multi-value search of a field in the table. The search endpoint should be configured in API Gateway:

            POST /search/{search_key}

        The value of {search_key} should be passed in as `key` and the contents of the POST request body should be
        passed in as the `values` list.

        :param Request request: Request object
        :param str key: Field to search
        :param list of str values: Search values
        :return: Search results
        :rtype: dict
        """
        # Get user
        user = self.initialize_request(request)

        # Build multi-value filter expressions
        expressions = self.filtering.multi_filter(user, key, values)
        sentry_sdk.add_breadcrumb(category='data', message='Built multi-value filter', level='info')

        # Perform each generated filter expression and then combine the results together
        output = []
        for expression in expressions:
            # Download data
            data = self._scan(self.data_table, FilterExpression=expression)

            sentry_sdk.add_breadcrumb(
                category='query', message=expression, level='info', table=self.config.data_table
            )

            # Add to output
            output.extend(data)

        # Create audit log
        self.audit_log(action=self.AUDIT_ACTION_SEARCH, request=request, user=user)

        # Return the filtered response
        return self.post_process(output, user)

    def delete(self, request: Request, primary_key: dict, condition=None, condition_failure_message='') -> dict:
        """
        Delete an item

        :param Request request: Request object
        :param dict primary_key: Dictionary formatted as {"primary_key": "value_of_row_to_delete"}
        :param boto3.dynamodb.conditions.ComparisonCondition condition: Optional condition to apply to this deletion.
        :param str condition_failure_message: If the conditional check fails, this optional error message
        will be displayed
        to the user.
        :return: Success
        :rtype: dict
        """
        # Get user
        user = self.initialize_request(request)

        # Build default conditional failure message
        if condition:
            default_condition_message = 'Conditional check failed'
        else:
            default_condition_message = 'Item does not exist or you do not have permission to delete it'

        # Add in the user's permissions
        conditions = self.filtering.filter(user, action=self.filtering.FILTER_ACTION_DELETE)

        # Add in the conditional expression from the user
        if condition:
            conditions = self.filtering.And(conditions, condition)

        # Default conditional expression to make sure the item actually exists
        for key, value in primary_key.items():
            conditions = self.filtering.And(conditions, Attr(key).eq(value))

        # Perform the deletion
        try:
            self.data_table.delete_item(Key=primary_key, ConditionExpression=conditions)
            logger.info('[%(user)s] Successfully deleted record "%(primary_key)s"' % {
                'user': self.user_identifier(user),
                'primary_key': primary_key,
            })
        except ClientError as e:
            if e.response['Error']['Code'] == 'ConditionalCheckFailedException':
                logger.warning(
                    '[%(user)s] Unable to delete record "%(primary_key)s" because the conditional check failed' % {
                        'user': self.user_identifier(user),
                        'primary_key': primary_key
                    }
                )
                raise BadRequestException(condition_failure_message or default_condition_message)
            logger.error(
                '[%(user)s] Encountered error while attempting to delete record '
                '"%(primary_key)s": [%(code)s] %(error)s' % {
                    'user': self.user_identifier(user),
                    'primary_key': primary_key,
                    'code': e.response['Error']['Code'],
                    'error': e.response['Error']['Message'],
                }
            )
            raise
        except Exception as e:
            logger.error(
                '[%(user)s] Encountered error while attempting to delete record "%(primary_key)s": %(error)s' % {
                    'user': self.user_identifier(user),
                    'primary_key': primary_key,
                    'error': str(e)
                }
            )
            raise

        sentry_sdk.add_breadcrumb(category='data', message='Deleted item', level='info')

        # Create audit log
        self.audit_log(
            action=self.AUDIT_ACTION_DELETE,
            request=request,
            user=user,
            resource=primary_key,
        )

        # Return updated record
        return {'deleted': True}



import os

import boto3

from scoutr.models.config import Config

try:
    import sentry_sdk
    from sentry_sdk.integrations.aws_lambda import AwsLambdaIntegration
    has_sentry = True
except ImportError:
    has_sentry = False


def configure_sentry():
    if not has_sentry:
        return

    secret_id = os.getenv('SENTRY_DSN_SECRET_ID')
    if not secret_id:
        return

    # Pull from secrets manager
    secrets_manager = boto3.client('secretsmanager')
    dsn = secrets_manager.get_secret_value(secret_id)['SecretString']
    if dsn:
        sentry_sdk.init(
            dsn=dsn,
            integrations=[AwsLambdaIntegration()],
            environment=os.getenv('Stage')
        )


def get_config():
    return Config(
        data_table=os.getenv('TableName'),
        auth_table=os.getenv('AuthTable'),
        group_table=os.getenv('GroupTable'),
        audit_table=os.getenv('AuditTable'),
        primary_key='id'
    )



import os

# import sentry_sdk
import firebase_admin
from firebase_admin import credentials, firestore
from flask import request
# from sentry_sdk.integrations.flask import FlaskIntegration
from scoutr.models.config import Config
from scoutr.providers.gcp.api import FirestoreAPI
from scoutr.helpers.flask.oidc import build_oidc_request
from scoutr.helpers.flask.routes import init_flask
from scoutr.helpers.flask.utils import flaskapi_exception_wrapper

# sentry_sdk.init(
#     dsn='https://APIKEY@sentry.io/PROJECT-ID',
#     integrations=[FlaskIntegration()],
#     environment=os.getenv('ENV', 'dev'),
#     release='PROJECT-NAME@VERSION'
# )

config = Config(
    data_table='data',
    auth_table='auth',
    group_table='groups',
    audit_table='audit',
    primary_key='id',
    oidc_username_header='Oidc-Claim-Sub',
    oidc_name_header='Oidc-Claim-Name',
    oidc_email_header='Oidc-Claim-Mail',
    oidc_group_header='Oidc-Claim-Groups'
)

# Initialize credentials
cred = credentials.Certificate('/home/scoutr/gcp.json')
firebase_admin.initialize_app(cred)

# Create Scoutr instance
api = FirestoreAPI(config)
app = init_flask(
    api=api,
    primary_list_endpoint='/items/',
)


@app.route("/item/", methods=['POST'])
@flaskapi_exception_wrapper
def create_item():
    """Create an item"""
    return api.create(
        request=build_oidc_request(api, request),
        data=request.data
    )


if __name__ == "__main__":
    app.run(debug=True)



from typing import Set


def merge_lists(primary_list: list, secondary_list: list) -> list:
    """Merge two lists together and return a unique list"""
    return primary_list + [item for item in secondary_list if item not in primary_list]


def value_in_set(value: str, valid_options: Set[str], option_name: str = 'option',
                 custom_error_message: str = ''):
    """
    Check if a value is contained in a list of valid options. This is supplied as a convenience function
    and is intended to be used with the input field validation on creates/updates.

    :param str value: Value to check
    :param set of str valid_options: List of options that the value should be included in for this to be successful
    :param str option_name: Optional descriptive name of the option that can be used to enrich an error message.
    :param str custom_error_message: Optional custom error message to return instead of the default one.
    :return: Dictionary that can be used with the field_validation
    :rtype: dict
    """
    return {
        'result': value in valid_options,
        'message': custom_error_message or f'{value} is not a valid {option_name}. Valid options: {valid_options}'
    }



class HttpException(Exception):
    status = 500

    def __init__(self, status=500, *args, **kwargs):
        self.status = status
        super(HttpException, self).__init__(*args, **kwargs)


class BadRequestException(HttpException):
    status = 400

    def __init__(self, *args, **kwargs):
        super(HttpException, self).__init__(*args, **kwargs)


class ForbiddenException(HttpException):
    status = 401

    def __init__(self, *args, **kwargs):
        super(HttpException, self).__init__(*args, **kwargs)


class UnauthorizedException(HttpException):
    status = 403

    def __init__(self, *args, **kwargs):
        super(HttpException, self).__init__(*args, **kwargs)


class NotFoundException(HttpException):
    status = 404

    def __init__(self, *args, **kwargs):
        super(HttpException, self).__init__(*args, **kwargs)


class InvalidUserException(BadRequestException):
    pass



import json
import re
from abc import abstractmethod
from decimal import Decimal
from typing import Any, Optional, List, Dict, Tuple, Callable, Set
from urllib.parse import unquote_plus

from scoutr.exceptions import BadRequestException
from scoutr.models.user import FilterField, User


class Filtering:

    OPERATION_EQUAL = 'eq'
    OPERATION_NOT_EQUAL = 'ne'
    OPERATION_STARTS_WITH = 'startswith'
    OPERATION_CONTAINS = 'contains'
    OPERATION_NOT_CONTAINS = 'notcontains'
    OPERATION_EXISTS = 'exists'
    OPERATION_GREATER_THAN = 'gt'
    OPERATION_LESS_THAN = 'lt'
    OPERATION_GREATER_THAN_EQUAL = 'ge'
    OPERATION_LESS_THAN_EQUAL = 'le'
    OPERATION_BETWEEN = 'between'
    OPERATION_IN = 'in'
    OPERATION_NOT_IN = 'notin'

    FILTER_ACTION_READ = 'READ'
    FILTER_ACTION_CREATE = 'CREATE'
    FILTER_ACTION_UPDATE = 'UPDATE'
    FILTER_ACTION_DELETE = 'DELETE'

    NUMERIC_OPERATIONS = (
        OPERATION_GREATER_THAN,
        OPERATION_LESS_THAN,
        OPERATION_GREATER_THAN_EQUAL,
        OPERATION_LESS_THAN_EQUAL,
    )

    @property
    def operations(self) -> Dict[str, Callable[[str, Any], Any]]:
        """
        List of supported operations

        :return: Supported operations
        :rtype: dict
        """
        return {
            self.OPERATION_EQUAL: self.equals,
            self.OPERATION_NOT_EQUAL: self.not_equal,
            self.OPERATION_CONTAINS: self.contains,
            self.OPERATION_NOT_CONTAINS: self.not_contains,
            self.OPERATION_STARTS_WITH: self.starts_with,
            self.OPERATION_EXISTS: self.exists,
            self.OPERATION_GREATER_THAN: self.greater_than,
            self.OPERATION_LESS_THAN: self.less_than,
            self.OPERATION_GREATER_THAN_EQUAL: self.greater_than_equal,
            self.OPERATION_LESS_THAN_EQUAL: self.less_than_equal,
            self.OPERATION_BETWEEN: self.between,
            self.OPERATION_IN: self.is_in,
            self.OPERATION_NOT_IN: self.not_in
        }

    @abstractmethod
    def And(self, condition1, condition2):
        raise NotImplementedError

    @abstractmethod
    def Or(self, condition1, condition2):
        raise NotImplementedError

    @abstractmethod
    def equals(self, attr: str, value):
        raise NotImplementedError

    @abstractmethod
    def not_equal(self, attr: str, value):
        raise NotImplementedError

    @abstractmethod
    def starts_with(self, attr: str, value):
        raise NotImplementedError

    @abstractmethod
    def contains(self, attr: str, value):
        raise NotImplementedError

    @abstractmethod
    def not_contains(self, attr: str, value):
        raise NotImplementedError

    @abstractmethod
    def exists(self, attr: str, value):
        raise NotImplementedError

    @abstractmethod
    def greater_than(self, attr: str, value):
        raise NotImplementedError

    @abstractmethod
    def less_than(self, attr: str, value):
        raise NotImplementedError

    @abstractmethod
    def greater_than_equal(self, attr: str, value):
        raise NotImplementedError

    @abstractmethod
    def less_than_equal(self, attr: str, value):
        raise NotImplementedError

    @abstractmethod
    def between(self, attr: str, value):
        raise NotImplementedError

    @abstractmethod
    def is_in(self, attr: str, value):
        raise NotImplementedError

    @abstractmethod
    def not_in(self, attr: str, value):
        raise NotImplementedError

    def _user_filters(self, filter_fields: List[FilterField]) -> Any:
        # Merge all possible values for this filter key together
        filters: Dict[str, List[FilterField]] = {}
        for filter_field in filter_fields:
            filters.setdefault(filter_field.field, list())
            filters[filter_field.field].append(filter_field)

        # Perform the filter
        conditions = None
        for key, filter_items in filters.items():
            if len(filter_items) == 1:
                # Perform a single query
                item = filter_items[0]
                conditions = self.perform_filter(conditions, f'{key}__{item.operator}', item.value)
            elif len(filter_items) > 1:
                # Perform an OR query against all possible values for this key
                filter_conds = None
                for item in filter_items:
                    filter_conds = self.Or(
                        filter_conds,
                        self.perform_filter(None, f'{key}__{item.operator}', item.value)
                    )
                conditions = self.And(conditions, filter_conds)

        return conditions

    def filter(self, user: Optional[User], filters: Optional[dict] = None, action: str = FILTER_ACTION_READ) -> Any:
        conditions = None
        if filters is None:
            filters = {}

        if user:
            # Select filter type (defaults to read filters)
            if action == self.FILTER_ACTION_CREATE:
                filter_fields = user.create_filters
            elif action == self.FILTER_ACTION_UPDATE:
                filter_fields = user.update_filters
            elif action == self.FILTER_ACTION_DELETE:
                filter_fields = user.delete_filters
            else:
                filter_fields = user.read_filters

            # Perform user filter
            conditions = self._user_filters(filter_fields)

        # Run filter operation
        return self._filter(conditions, filters)

    def _filter(self, conditions: Any, filters: dict) -> Any:
        # Build filters that were passed in
        for key, value in filters.items():
            if isinstance(value, list):
                # Multi-value filter
                for item in value:
                    if not isinstance(item, str):
                        raise BadRequestException('Query filter value must be a string or list of strings')
                    conditions = self.perform_filter(conditions, key, item)
            elif isinstance(value, str):
                # Single value filter
                conditions = self.perform_filter(conditions, key, value)
            else:
                # Invalid
                raise BadRequestException('Query filter value must be a string or list of strings')

        return conditions

    def get_operator(self, key: str) -> Tuple[str, str]:
        # Check if this is a magic operator
        operation = self.OPERATION_EQUAL
        magic_operator_match = re.match('^(.+)__(.+)$', key)
        if magic_operator_match:
            key = magic_operator_match.group(1)
            operation = magic_operator_match.group(2)

        # Fetch condition function from operation map
        return key, operation

    def perform_filter(self, conditions: Any, key: str, value: Any) -> Any:
        # Unquote the value
        if isinstance(value, str):
            value = unquote_plus(value)
            if value == '':
                raise BadRequestException('Filter key %s has no value' % key)

        # Get operator
        key, operator = self.get_operator(key)

        # Convert to decimal if this is a numeric operation
        if isinstance(value, str) and value.isnumeric() and operator in self.NUMERIC_OPERATIONS:
            value = Decimal(value)

        # Perform the filter operation
        func = self.operations.get(operator)
        try:
            # Check if this is an unknown operation
            if func is None:
                raise NotImplementedError

            # Run the condition function
            condition = func(key, value)

            # If condition is null, do not apply the condition
            if condition is not None:
                return self.And(conditions, condition)
        except NotImplementedError:
            raise BadRequestException(f"Provider does not support magic operator '{operator}")

        return conditions

    def multi_filter(self, user, filter_key, value):
        # Make sure a value was provided
        if not value:
            raise BadRequestException('No search values were provided')

        # Build pre-set filters
        base_condition = self.filter(user)

        # Build the multi-filters
        expressions = []

        if isinstance(value, list):
            if len(value) < 99:
                condition = self.And(base_condition, self.is_in(filter_key, value))
                expressions.append(condition)
            else:
                for i in range(0, len(value), 99):
                    condition = self.And(base_condition, self.is_in(filter_key, value[i:i + 99]))
                    expressions.append(condition)
        else:
            condition = self.And(base_condition, self.equals(filter_key, value))
            expressions.append(condition)

        return expressions


class LocalFiltering(Filtering):
    def __init__(self, data: dict):
        self.data = data
        self.failed_filters: Set[str] = set()

    def _user_filters(self, filter_fields: List[FilterField]) -> Any:
        # Merge all possible values for this filter key together
        filters: Dict[str, List[FilterField]] = {}
        for filter_field in filter_fields:
            filters.setdefault(filter_field.field, list())
            filters[filter_field.field].append(filter_field)

        # Perform the filter
        conditions = None
        for key, filter_items in filters.items():
            if len(filter_items) == 1:
                # Perform a single query
                item = filter_items[0]
                existing_conditions = conditions
                conditions = self.perform_filter(conditions, f'{key}__{item.operator}', item.value)
                if (existing_conditions is None or existing_conditions is True) and conditions is False:
                    self.failed_filters.add(key)
            elif len(filter_items) > 1:
                # Perform an OR query against all possible values for this key
                filter_conds = None
                for item in filter_items:
                    filter_conds = self.Or(
                        filter_conds,
                        self.perform_filter(None, f'{key}__{item.operator}', item.value)
                    )
                existing_conditions = conditions
                conditions = self.And(conditions, filter_conds)
                if (existing_conditions is None or existing_conditions is True) and conditions is False:
                    self.failed_filters.add(key)

        return conditions

    def And(self, condition1, condition2):
        if condition1 is None:
            condition1 = True
        if condition2 is None:
            condition2 = True

        return condition1 and condition2

    def Or(self, condition1, condition2):
        if condition1 is None:
            condition1 = True
        if condition2 is None:
            condition2 = True

        return condition1 or condition2

    def equals(self, attr: str, value):
        if attr not in self.data:
            return False
        return self.data[attr] == value

    def not_equal(self, attr: str, value):
        if attr not in self.data:
            return False
        return self.data[attr] != value

    def contains(self, attr: str, value: str):
        if attr not in self.data:
            return False
        return value in self.data[attr]

    def not_contains(self, attr: str, value: str):
        if attr not in self.data:
            return False
        return value not in self.data[attr]

    def starts_with(self, attr: str, value: str):
        if attr not in self.data:
            return False
        return self.data[attr].startswith(value)

    def exists(self, attr: str, value):
        if value == 'true':
            return attr in self.data
        elif value == 'false':
            return attr not in self.data
        else:
            raise Exception('Invalid value for exists operation')

    def greater_than(self, attr: str, value):
        if attr not in self.data:
            return False
        return self.data[attr] > value

    def less_than(self, attr: str, value):
        if attr not in self.data:
            return False
        return self.data[attr] < value

    def greater_than_equal(self, attr: str, value):
        if attr not in self.data:
            return False
        return self.data[attr] >= value

    def less_than_equal(self, attr: str, value):
        if attr not in self.data:
            return False
        return self.data[attr] <= value

    def between(self, attr: str, value):
        if attr not in self.data:
            return False
        if isinstance(value, list):
            values = value
        else:
            values = json.loads(value)
        if not len(values) == 2:
            raise Exception('Between operation requires two values')
        return values[0] <= self.data[attr] <= values[1]

    def is_in(self, attr: str, value):
        if attr not in self.data:
            return False
        if isinstance(value, list):
            values = value
        else:
            values = json.loads(value)
        return self.data[attr] in values

    def not_in(self, attr: str, value):
        if attr not in self.data:
            return False
        if isinstance(value, list):
            values = value
        else:
            values = json.loads(value)
        return self.data[attr] not in values



import json
from typing import Dict, Callable, Any

from scoutr.providers.base.filtering import Filtering


class MongoFiltering(Filtering):
    OPERATION_REGEX = 'regex'
    OPERATION_TYPE = 'type'
    OPERATION_HAS_ELEMENTS = 'haselements'

    @property
    def operations(self) -> Dict[str, Callable[[str, Any], Any]]:
        ops = super(MongoFiltering, self).operations
        ops.update({
            self.OPERATION_REGEX: self.regex,
            self.OPERATION_TYPE: self.is_type,
            self.OPERATION_HAS_ELEMENTS: self.has_elements
        })
        return ops

    def And(self, condition1, condition2):
        if condition1 and condition2:
            return {'$and': [condition1, condition2]}
        elif condition1:
            return condition1
        elif condition2:
            return condition2
        else:
            return None

    def Or(self, condition1, condition2):
        if condition1 and condition2:
            return {'$or': [condition1, condition2]}
        elif condition1:
            return condition1
        elif condition2:
            return condition2
        else:
            return None

    def equals(self, attr: str, value):
        return {
            attr: {'$eq': value}
        }

    def not_equal(self, attr: str, value):
        return {
            attr: {'$ne': value}
        }

    def contains(self, attr: str, value):
        return {
            attr: {'$regex': f'.*{value}.*'}
        }

    def not_contains(self, attr: str, value):
        return {
            attr: {'$regex': f'^(?:(?!{value}).)*$'}
        }

    def starts_with(self, attr: str, value):
        return {
            attr: {'$regex': f'^{value}.*'}
        }

    @staticmethod
    def has_elements(attr: str, value):
        values = json.loads(value)

        if isinstance(values, list):
            values = {
                '$in': values
            }
        elif not isinstance(values, dict):
            values = {
                '$eq': str(values)
            }
        return {
            attr: {'$elemMatch': values}
        }

    def greater_than(self, attr: str, value):
        return {
            attr: {'$gt': value}
        }

    def greater_than_equal(self, attr: str, value):
        return {
            attr: {'$gte': value}
        }

    def less_than(self, attr: str, value):
        return {
            attr: {'$lt': value}
        }

    def less_than_equal(self, attr: str, value):
        return {
            attr: {'$lte': value}
        }

    def between(self, attr: str, value):
        if isinstance(value, list):
            values = value
        else:
            values = json.loads(value)
        if not len(values) == 2:
            raise Exception('Between operation requires two values')
        return {
            attr: {
                '$gte': values[0],
                '$lte': values[1]
            }
        }

    def is_in(self, attr: str, value):
        if isinstance(value, list):
            values = value
        else:
            values = json.loads(value)

        return {
            attr: {'$in': values}
        }

    def not_in(self, attr: str, value):
        if isinstance(value, list):
            values = value
        else:
            values = json.loads(value)

        return {
            attr: {'$nin': values}
        }

    def exists(self, attr: str, value):
        if value == "true":
            return {attr: {'$exists': True}}
        elif value == "false":
            return {attr: {'$exists': False}}
        else:
            raise Exception('Invalid value for exists operation')

    @staticmethod
    def is_type(attr: str, value):
        return {
            attr: {'$type': value}
        }

    @staticmethod
    def regex(attr: str, value):
        return {
            attr: {'$regex': value}
        }



import json

from boto3.dynamodb.conditions import Attr, Not

from scoutr.providers.base.filtering import Filtering


class AWSFiltering(Filtering):
    def And(self, condition1, condition2):
        if condition1 and condition2:
            return condition1 & condition2
        elif condition1:
            return condition1
        elif condition2:
            return condition2
        else:
            return None

    def Or(self, condition1, condition2):
        if condition1 and condition2:
            return condition1 | condition2
        elif condition1:
            return condition1
        elif condition2:
            return condition2
        else:
            return None

    def equals(self, attr: str, value):
        return Attr(attr).eq(value)

    def not_equal(self, attr: str, value):
        return Attr(attr).ne(value)

    def contains(self, attr: str, value):
        return Attr(attr).contains(value)

    def not_contains(self, attr: str, value):
        return Not(Attr(attr).contains(value))

    def starts_with(self, attr: str, value):
        return Attr(attr).begins_with(value)

    def exists(self, attr: str, value):
        if value == "true":
            return Attr(attr).exists()
        elif value == "false":
            return Attr(attr).not_exists()
        else:
            raise Exception('Invalid value for exists operation')

    def greater_than(self, attr: str, value):
        return Attr(attr).gt(value)

    def less_than(self, attr: str, value):
        return Attr(attr).lt(value)

    def greater_than_equal(self, attr: str, value):
        return Attr(attr).gte(value)

    def less_than_equal(self, attr: str, value):
        return Attr(attr).lte(value)

    def between(self, attr: str, value):
        if isinstance(value, list):
            values = value
        else:
            values = json.loads(value)
        if not len(values) == 2:
            raise Exception('Between operation requires two values')
        return Attr(attr).between(*values)

    def is_in(self, attr: str, value):
        if isinstance(value, list):
            values = value
        else:
            values = json.loads(value)
        return Attr(attr).is_in(values)

    def not_in(self, attr: str, value):
        if isinstance(value, list):
            values = value
        else:
            values = json.loads(value)
        return Not(Attr(attr).is_in(values))



import os
from typing import List

from flask_api.request import APIRequest

from scoutr.exceptions import UnauthorizedException
from scoutr.models.request import Request, UserData, RequestUser
from scoutr.providers.base.api import BaseAPI


def get_user_from_oidc(api: BaseAPI, request: APIRequest) -> RequestUser:
    groups: List[str] = []

    # Return a dummy user when in debug mode
    if os.getenv('DEBUG', 'false') == 'true':
        group_string = os.getenv('GROUPS', '')
        if group_string:
            groups = group_string.split(',')

        return RequestUser(
            id='222222222',
            data=UserData(
                username='222222222',
                email='george.p.burdell@ge.com',
                name='Georgia Burdell',
                groups=groups
            )
        )

    if api.config.oidc_group_header:
        group_string = request.headers.get(api.config.oidc_group_header, '')
        if group_string == '':
            groups = []
        else:
            groups = group_string.split(',')

    # Permit tuples for the name header in case the name is split into first and last name
    if isinstance(api.config.oidc_name_header, tuple):
        name_parts = []
        for header in api.config.oidc_name_header:
            value = request.headers.get(header)
            if header:
                name_parts.append(value)
        name = ' '.join([
            request.headers.get(header)
            for header in api.config.oidc_name_header
            if request.headers.get(header)
        ]) or None
    else:
        name = request.headers.get(api.config.oidc_name_header)

    username = request.headers.get(api.config.oidc_username_header)
    if not username:
        raise UnauthorizedException('Unable to determine user id')

    user_data = UserData(
        name=name,
        email=request.headers.get(api.config.oidc_username_header),
        username=request.headers.get(api.config.oidc_username_header),
        groups=groups
    )

    return RequestUser(
        id=user_data.username,
        data=user_data
    )


def build_oidc_request(api: BaseAPI, request: APIRequest) -> Request:
    from scoutr.helpers.flask.utils import parse_query_params

    req = Request(
        method=request.method,
        path=request.path,
        source_ip=request.remote_addr,
        user_agent=request.user_agent.string,
        user=get_user_from_oidc(api, request)
    )

    query_params = parse_query_params(request.query_string)
    if query_params:
        req.query_params = query_params

    if request.data:
        req.body = request.data

    return req



from typing import Tuple

import simplejson
from flask import request
from flask_api import FlaskAPI
from flask_api.exceptions import ParseError
from werkzeug.middleware.proxy_fix import ProxyFix

from scoutr.helpers.flask.oidc import build_oidc_request
from scoutr.helpers.flask.utils import init_flask_user, flaskapi_exception_wrapper
from scoutr.providers.base.api import BaseAPI


def init_flask(api: BaseAPI, primary_list_endpoint: str, history_actions: Tuple[str] = ('CREATE', 'UPDATE')):
    """
    Initialize flask app

    :param scoutr.providers.base.api.BaseAPI api:
    :param str primary_list_endpoint:
    :param tuple of str history_actions:
    :return: Flask App
    :rtype: flask_api.FlaskAPI
    """
    # Make sure primary list endpoint is formatted as /endpoint/
    if not primary_list_endpoint.startswith('/'):
        primary_list_endpoint = '/' + primary_list_endpoint
    if not primary_list_endpoint.endswith('/'):
        primary_list_endpoint += '/'

    # Initialize the app
    app = FlaskAPI(__name__)
    app.wsgi_app = ProxyFix(app.wsgi_app)
    app.json_encoder = simplejson.JSONEncoder

    @app.route('/user/', methods=['GET'])
    @flaskapi_exception_wrapper
    def get_user():
        user = init_flask_user(api, request)
        return user.dict()

    @app.route('/user/has-permission/', methods=['POST'])
    @flaskapi_exception_wrapper
    def check_user_permissions():
        if 'method' not in request.data or 'path' not in request.data:
            raise ParseError("Body should contain 'method' and 'path' keys")

        # Build OIDC request
        req = build_oidc_request(api, request)

        return {
            'authorized': api.can_access_endpoint(
                method=request.data['method'],
                path=request.data['path'],
                user=req.user,
                request=req
            )
        }

    @app.route(primary_list_endpoint, methods=['GET'])
    @flaskapi_exception_wrapper
    def list_items():
        return api.list(request=build_oidc_request(api, request))

    @app.route('/audit/', methods=['GET'], defaults={'item': None})
    @app.route('/audit/<item>/', methods=['GET'])
    @flaskapi_exception_wrapper
    def audit(item):
        path_params = {}
        if item:
            path_params = {f'resource.{api.config.primary_key}': item}
        return api.list_audit_logs(
            request=build_oidc_request(api, request),
            param_overrides=path_params
        )

    @app.route('/history/<item>/', methods=['GET'])
    @flaskapi_exception_wrapper
    def history(item):
        return api.history(
            request=build_oidc_request(api, request),
            key=api.config.primary_key,
            value=item,
            actions=history_actions
        )

    @app.route('/search/<search_key>/', methods=['POST'])
    @flaskapi_exception_wrapper
    def search(search_key):
        return api.search(
            request=build_oidc_request(api, request),
            key=search_key,
            values=request.data
        )

    return app



import json

from scoutr.exceptions import HttpException
from scoutr.models.request import Request, RequestUser


def build_api_gateway_request(event: dict) -> Request:
    request = Request(
        method=event['httpMethod'],
        path=event['path'],
        source_ip=event['requestContext']['identity']['sourceIp'],
        user_agent=event['requestContext']['identity']['userAgent'],
        user=get_api_gateway_user(event)
    )

    query_params = event.get('multiValueQueryStringParameters', {}) or {}
    if query_params:
        request.query_params = query_params

    path_params = event.get('pathParameters', {}) or {}
    if path_params:
        request.path_params = path_params

    body = event.get('body', '{}') or '{}'
    if not isinstance(body, dict):
        try:
            body = json.loads(body)
        except json.JSONDecodeError:
            pass
    if body:
        request.body = body

    return request


def get_api_gateway_user(event: dict) -> RequestUser:
    return RequestUser(id=event['requestContext']['identity']['apiKeyId'], data=None)


def handle_http_exception(e: HttpException):
    if len(e.args) == 1 and isinstance(e.args[0], (list, dict)):
        return {
            'statusCode': e.status,
            'body': json.dumps({'errors': e.args[0]})
        }

    return {
        'statusCode': e.status,
        'body': json.dumps({'error': str(e)})
    }



from typing import Dict, Union, List, Optional

from scoutr.models import Model
from scoutr.models.user import FilterField


class AuditUser(Model):
    id: str
    username: str
    name: str
    email: str
    source_ip: str
    user_agent: str
    read_filters: List[FilterField]
    create_filters: List[FilterField]
    update_filters: List[FilterField]
    delete_filters: List[FilterField]


class AuditLog(Model):
    time: str
    user: AuditUser
    action: str
    method: str
    path: str
    expire_time: Optional[int]
    query_params: Optional[Dict[str, str]]
    resource: Optional[Dict[str, str]]
    body: Optional[Union[dict, list]]



import json
import re
from abc import abstractmethod
from concurrent import futures
from concurrent.futures.thread import ThreadPoolExecutor
from copy import deepcopy
from datetime import datetime, timedelta
from typing import List, Dict, Optional, Any, Callable, Union, Tuple, Set

from scoutr.exceptions import UnauthorizedException, BadRequestException, ForbiddenException, NotFoundException
from scoutr.models.audit import AuditLog, AuditUser
from scoutr.models.config import Config
from scoutr.models.request import Request, UserData
from scoutr.models.user import User, Group, Permissions
from scoutr.providers.base.filtering import Filtering, LocalFiltering
from scoutr.utils.utils import merge_lists

try:
    import sentry_sdk
except ImportError:
    from scoutr.utils import mock_sentry
    sentry_sdk = mock_sentry


class BaseAPI:
    filtering: Filtering
    unique_func: Callable[[List[Dict], str], List[str]] = \
        lambda data, key: sorted(set([item[key] for item in data if item and item[key]]))

    AUDIT_ACTION_CREATE = 'CREATE'
    AUDIT_ACTION_UPDATE = 'UPDATE'
    AUDIT_ACTION_LIST = 'LIST'
    AUDIT_ACTION_GET = 'GET'
    AUDIT_ACTION_SEARCH = 'SEARCH'
    AUDIT_ACTION_DELETE = 'DELETE'

    def __init__(self, config: Config):
        self.config = config

    def can_access_endpoint(self, method: str, path: str, user: User, request: Request = None) -> bool:
        if request:
            # Fetch the user
            try:
                user = self.get_user(request.user.id, request.user.data)
            except Exception as e:
                print(f'Failed to fetch user: {e}')
                return False

            # Validate the user
            try:
                self.validate_user(user)
            except Exception as e:
                print(f'Encountered error while validating user: {e}')
                return False

        # Verify user was provided/looked up
        if user is None:
            print('Unable to validate if user has access to endpoint because user was None')
            return False

        # Check permitted endpoints
        for item in user.permitted_endpoints:
            if method == item.method and re.match(item.endpoint, path):
                return True

        return False

    def initialize_request(self, request: Request) -> User:
        # Get user
        user = self.get_user(request.user.id, request.user.data)

        # Validate user
        self.validate_user(user)

        # Validate request
        self.validate_request(request, user)

        return user

    @staticmethod
    def merge_permissions(user: User, group: Group):
        for attribute in Permissions.attributes():
            value = merge_lists(getattr(user, attribute), getattr(group, attribute))
            setattr(user, attribute, value)

    def get_user(self, user_id: str, user_data: Optional[UserData] = None) -> User:
        is_user = True
        user = User(id=user_id)

        # Try to find user in the auth table
        auth = self.get_auth(user_id)
        if not auth:
            # Failed to find user in the table
            is_user = False
        else:
            user = auth
            user.id = user_id

        # Try to find supplied entitlements in the auth table
        entitlement_ids: List[str] = []
        if user_data and user_data.groups:
            entitlements = self.get_entitlements(user_data.groups)
            for entitlement in entitlements:
                # Store this as a real entitlement
                entitlement_ids.append(entitlement.id)

                # Add sub-groups
                user.groups.extend(entitlement.groups)

                # Merge permissions
                user.permitted_endpoints = merge_lists(user.permitted_endpoints, entitlement.permitted_endpoints)
                user.exclude_fields = merge_lists(user.exclude_fields, entitlement.exclude_fields)
                user.update_fields_restricted = merge_lists(
                    user.update_fields_restricted,
                    entitlement.update_fields_restricted
                )
                user.update_fields_permitted = merge_lists(
                    user.update_fields_permitted,
                    entitlement.update_fields_permitted
                )
                user.read_filters = merge_lists(user.read_filters, entitlement.read_filters)
                user.create_filters = merge_lists(user.create_filters, entitlement.create_filters)
                user.update_filters = merge_lists(user.update_filters, entitlement.update_filters)
                user.delete_filters = merge_lists(user.delete_filters, entitlement.delete_filters)

        # Check that a user was found
        if not is_user and not entitlement_ids:
            raise UnauthorizedException(f"Auth id '{user_id}' is not authorized")

        # If the user is a member of a group, merge in the group's permissions
        for group_id in user.groups:
            group = self.get_group(group_id)
            if not group:
                # Group is not in the table
                raise UnauthorizedException(f"Group '{group_id}' does not exist")

            # Merge user and group permissions together
            self.merge_permissions(user, group)

        # Save user groups before applying metadata
        user_groups = user.groups.copy()

        # Update user object with metadata
        if user_data:
            if user_data.username:
                user.username = user_data.username
            if user_data.name:
                user.name = user_data.name
            if user_data.email:
                user.email = user_data.email
            if user_data.groups:
                user.groups = user_data.groups

        # Update user object with all applied entitlements
        if entitlement_ids:
            groups: List[str] = []
            groups.extend(user_groups)
            groups.extend(entitlement_ids)
            user.groups = groups

        return user

    @staticmethod
    def validate_user(user: User):
        # Make sure the user contains the required keys
        if not user.id or not user.username or not user.name or not user.email:
            raise UnauthorizedException('User missing one of the following fields: id, username, name, email')

        # Validate exclude fields
        for item in user.exclude_fields:
            if not isinstance(item, str):
                raise UnauthorizedException(f"User '{user.id}' field 'exclude_fields' must be a list of strings")

        # Validate read filters
        for read_filter in user.read_filters:
            if not read_filter.field or not read_filter.value:
                raise UnauthorizedException(
                    f"User '{user.id}' field 'read_filters' must be a list of dictionaries with each "
                    "item formatted as {'field': 'field_name', 'operation': 'eq', 'value': 'value'}"
                )

        # Validate create filters
        for create_filters in user.create_filters:
            if not create_filters.field or not create_filters.value:
                raise UnauthorizedException(
                    f"User '{user.id}' field 'create_filters' must be a list of dictionaries with each "
                    "item formatted as {'field': 'field_name', 'operation': 'eq', 'value': 'value'}"
                )

        # Validate update filters
        for update_filters in user.update_filters:
            if not update_filters.field or not update_filters.value:
                raise UnauthorizedException(
                    f"User '{user.id}' field 'update_filters' must be a list of dictionaries with each "
                    "item formatted as {'field': 'field_name', 'operation': 'eq', 'value': 'value'}"
                )

        # Validate delete filters
        for delete_filters in user.delete_filters:
            if not delete_filters.field or not delete_filters.value:
                raise UnauthorizedException(
                    f"User '{user.id}' field 'delete_filters' must be a list of dictionaries with each "
                    "item formatted as {'field': 'field_name', 'operation': 'eq', 'value': 'value'}"
                )

        # Make sure all the endpoints are valid regex
        for permitted_endpoint in user.permitted_endpoints:
            try:
                re.compile(permitted_endpoint.endpoint)
            except Exception as e:
                raise BadRequestException(f'Failed to compile endpoint regex: {e}')

    def validate_request(self, req: Request, user: User):
        # Make sure the user has permissions to access this endpoint
        if self.can_access_endpoint(req.method, req.path, user):
            # Log request
            user_id = self.user_identifier(user)
            if req.method == 'GET':
                print(f'[{user_id}] Performed {req.method} on {req.path}')
            else:
                print(f'[{user_id}] Performed {req.method} on {req.path}:\n{req.body}')

            # User is authorized to access this endpoint
            return

        # Make sure query params have keys and values
        if set(req.query_params.keys()).intersection(['']) or set(req.query_params.values()).intersection(['']):
            raise BadRequestException('Query strings must have keys and values')

        raise ForbiddenException(f'Not authorized to perform {req.method} on endpoint {req.path}')

    @staticmethod
    def validate_fields(validation: Optional[dict], required_fields: Union[List, Tuple], item: dict,
                        existing_item: Optional[dict] = None):
        """
        Perform field validation before creating/updating items into Dynamo

        Expected that the field_validation argument consists of a dictionary formatted as:

            {
                "field_name": callable_that_returns_bool,
                "field_name_2": callable_that_returns_dict
            }

            The callable should accept three arguments. The first is the input value and the second is the entire
            item that is being validating (in case complex filters are needed that cross-reference against other
            field values). The last argument is the value of the existing item (used during record updates). When
            validating record creations, the passed value will be `None`.

            It can either return a boolean or a dictionary. If it returns a boolean that evaluates to
            False, then a BadRequestException will be raised with the default error message "Invalid value for key
            'field_name'". If the callable returns a dictionary, it is expected that the dictionary is structured as:

                {
                    "result": bool,
                    "message": "message that should be returned to the user if 'result' is False"
                }

            This dictionary format allows for customizing the error message returned to the user.

        :param dict validation: Dictionary that maps fields to their validation callables
        :param list required_fields: List of fields that should be required
        :param dict item: Data that should be validated
        :param dict existing_item: Existing item in Dynamo
        :raises: BadRequestException
        """
        if validation is None:
            validation = {}

        # Check for required fields
        if required_fields:
            missing_keys = set(required_fields).difference(set(item.keys()))
            if missing_keys:
                raise BadRequestException(f'Missing required fields {missing_keys}')

        sentry_sdk.add_breadcrumb(category='validate', message='Validated required fields were included', level='info')

        # Perform field validation
        errors: Dict[str, str] = {}
        with ThreadPoolExecutor() as executor:
            threads = {}
            for key, func in validation.items():
                if key in item:
                    future = executor.submit(func, item[key], item, existing_item)
                    threads.update({future: key})

            for future in futures.as_completed(threads):
                key = threads[future]
                try:
                    result = future.result()
                except Exception as e:
                    errors[key] = str(e)
                    continue

                if isinstance(result, dict):
                    if 'result' not in result:
                        errors[key] = 'Validator is not properly configured'
                    elif not result['result']:
                        if not isinstance(result.get('message'), str):
                            errors[key] = 'Validator is not properly configured'
                            continue
                        errors[key] = result['message']
                elif not result:
                    errors[key] = 'Invalid value'

        if errors:
            raise BadRequestException(errors)

        sentry_sdk.add_breadcrumb(category='validate', message='Validated input fields', level='info')

    def _prepare_create(self, request: Request, data: dict, validation: dict = None,
                        required_fields: Union[List, Tuple] = ()) -> User:
        # Get user
        user = self.initialize_request(request)

        # Make sure the user has permission to update all of the fields specified
        unauthorized_fields = set(data.keys()).intersection(set(user.exclude_fields))
        if unauthorized_fields:
            raise UnauthorizedException(f'Not authorized to create item with fields {unauthorized_fields}')

        # Run validation
        self.validate_fields(validation, required_fields, data)

        # Creation filters
        local_filter = LocalFiltering(data)
        if local_filter.filter(user, action=local_filter.FILTER_ACTION_CREATE) is False:
            raise UnauthorizedException(f'Unauthorized value(s) for field(s): {local_filter.failed_filters}')

        return user

    def _prepare_list(self, request: Request, process_path_params: bool = True,
                      process_search_keys: bool = True) -> Tuple[User, Dict[str, str]]:
        # Get user
        user = self.initialize_request(request)

        # Build params
        params: Dict[str, str] = {}
        params.update(request.query_params)
        if process_path_params:
            params.update(request.path_params)

        # Generate dynamo search
        if process_search_keys:
            search_key = request.path_params.get('search_key')
            search_value = request.path_params.get('search_value')
            if search_key and search_value:
                # Map the search key and value into params
                params[search_key] = search_value
                del params['search_key']
                del params['search_value']

        return user, params

    @staticmethod
    def validate_update(user: User, data: dict, existing_item: dict):
        """
        Make sure the user has permission to update all of the fields specified

        :param User user: User object
        :param dict data: Data object
        :param dict existing_item: Existing object
        :raises UnauthorizedException
        """
        if user.update_fields_permitted:
            # User is only permitted to update specified list of fields. Determine list of fields the user
            # is not authorized to update
            unauthorized_fields = set(data.keys()).difference(set(user.update_fields_permitted))
            if unauthorized_fields:
                raise UnauthorizedException(f'Not authorized to update fields {unauthorized_fields}')

        # Combine both exclude_fields and update_fields_restricted together. If the user can't see the field, they don't
        # have permission to update the field.
        restricted_fields = set(user.exclude_fields + user.update_fields_restricted)
        if restricted_fields:
            # User is restricted from updating certain fields. Determine list of fields the user
            # is not authorized to update.
            unauthorized_fields = set(data.keys()).intersection(restricted_fields)
            if unauthorized_fields:
                raise UnauthorizedException(f'Not authorized to update fields {unauthorized_fields}')

        combined_object = existing_item.copy()
        combined_object.update(data)

        # Update filters
        local_filter = LocalFiltering(combined_object)
        if local_filter.filter(user, action=local_filter.FILTER_ACTION_UPDATE) is False:
            raise UnauthorizedException(f'Unauthorized value(s) for field(s): {local_filter.failed_filters}')

    def post_process(self, data: List[dict], user: User) -> List[dict]:
        # If no filtering is necessary, return output
        if not user.exclude_fields:
            return data

        fields_logged = set()
        for item in data:
            # Remove fields the user shouldn't be able to see
            for field in user.exclude_fields:
                if field in item:
                    if field not in fields_logged:
                        print(f'[{self.user_identifier(user)}] Excluding field "{field}" from response"')
                        fields_logged.add(field)
                    del item[field]

        # Return the filtered output
        return data

    def audit_log(self, action: str, request: Request, user: User,
                  resource: Dict[str, str] = None, changes: Dict[str, str] = None):
        # Only send audit logs if the table is configured
        if not self.config.audit_table:
            return

        # Create audit log
        now = datetime.utcnow()
        audit_log: AuditLog = AuditLog(
            time=now.isoformat(),
            user=AuditUser(
                id=user.id,
                name=user.name,
                email=user.email,
                username=user.username,
                source_ip=request.source_ip,
                user_agent=request.user_agent,
                read_filters=user.read_filters,
                create_filters=user.create_filters,
                update_filters=user.update_filters,
                delete_filters=user.delete_filters
            ),
            action=action,
            method=request.method,
            path=request.path,
            query_params=request.query_params,
            body=request.body
        )

        # Add expiry time for read events
        if action in (self.AUDIT_ACTION_GET, self.AUDIT_ACTION_LIST, self.AUDIT_ACTION_SEARCH):
            expire_time = now + timedelta(days=self.config.log_retention_days)
            audit_log.expire_time = int(expire_time.timestamp())

        # Add body
        if changes:
            audit_log.body = changes

        # Add resource
        if resource:
            audit_log.resource = resource

        # Marshal to dict
        item = audit_log.dict()

        # Add the record
        if not self.store_item(self.config.audit_table, item):
            print('Failed to store audit log')
            print('Failed audit log', item)

    @staticmethod
    def user_identifier(user: User):
        return f'{user.id}: {user.name} ({user.username} - {user.email})'

    @abstractmethod
    def store_item(self, table: str, item: dict) -> bool:
        raise NotImplementedError

    @abstractmethod
    def get_auth(self, user_id: str) -> Optional[User]:
        raise NotImplementedError

    @abstractmethod
    def get_entitlements(self, entitlement_ids: List[str]) -> List[User]:
        raise NotImplementedError

    @abstractmethod
    def get_group(self, group_id: str) -> Optional[Group]:
        raise NotImplementedError

    @abstractmethod
    def create(self, request: Request, data: dict, validation: dict,
               required_fields: Union[List, Tuple]) -> dict:
        raise NotImplementedError

    @abstractmethod
    def update(self, request: Request, partition_key: dict, data: dict, validation: dict) -> dict:
        raise NotImplementedError

    @abstractmethod
    def get(self, request: Request, key: Any, value: Any) -> dict:
        raise NotImplementedError

    @abstractmethod
    def list(self, request: Request) -> List[dict]:
        raise NotImplementedError

    @abstractmethod
    def list_unique_values(self, request: Request, key: str,
                           unique_func: Callable[[List, str], List[str]]) -> List[str]:
        raise NotImplementedError

    @abstractmethod
    def list_audit_logs(self, request: Request, param_overrides: dict) -> List[dict]:
        raise NotImplementedError

    def history(self, request: Request, key: str, value: str,
                actions: tuple = (AUDIT_ACTION_CREATE, AUDIT_ACTION_UPDATE, AUDIT_ACTION_DELETE)) -> List[dict]:
        """
        Given a resource key and value, build the history of the item over time.

        :param Request request: Request object
        :param str key: Resource key
        :param str value: Resource value
        :param tuple of str actions: List of actions to filter on
        :return: List of the record's history
        :rtype: list of dict
        """
        if not self.config.audit_table:
            raise NotFoundException('Audit logs are not enabled')

        param_overrides = {
            f'resource.{key}': value,
            'action__in': json.dumps(actions)
        }

        # Get user
        self.initialize_request(request)

        # Get the audit logs (reverse results so oldest item is first)
        logs = self.list_audit_logs(request, param_overrides)[::-1]

        # Check for no results
        if len(logs) == 0:
            return []

        # Get the original record
        current_item: dict = {'data': {}, 'time': None}
        for item in logs:
            if item['action'] == self.AUDIT_ACTION_CREATE:
                current_item['time'] = item['time']
                current_item['data'] = item['body']
                break

        # Build the record history
        history = [current_item]
        for item in logs:
            # Skip creation calls
            if item['action'] == self.AUDIT_ACTION_CREATE:
                continue
            elif item['action'] == self.AUDIT_ACTION_DELETE:
                history.insert(0, {'time': item['time'], 'data': {}})
                continue

            # Make a full copy of the current item
            current_item = deepcopy(current_item)

            # Update the current item with the changes
            current_item['time'] = item['time']
            current_item['data'].update(item.get('body', {}))

            # Insert at the top of the history (i.e. most recent first)
            history.insert(0, current_item)

        return history

    @abstractmethod
    def search(self, request: Request, key: str, values: List[str]) -> List[dict]:
        raise NotImplementedError

    @abstractmethod
    def delete(self, request: Request, partition_key: dict) -> dict:
        raise NotImplementedError



import logging
from typing import List, Dict, Optional, Any, Callable, Union, Tuple

from firebase_admin import firestore
from google.cloud.firestore_v1 import Client, DocumentSnapshot, CollectionReference, Query

from scoutr.exceptions import NotFoundException, BadRequestException
from scoutr.models.config import Config
from scoutr.models.request import Request
from scoutr.models.user import User, Group
from scoutr.providers.base.api import BaseAPI
from scoutr.providers.gcp.filtering import GCPFiltering

try:
    import sentry_sdk
    has_sentry = True
except ImportError:
    has_sentry = False

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class FirestoreAPI(BaseAPI):
    def __init__(self, config: Config):
        super(FirestoreAPI, self).__init__(config)
        self.db: Client = firestore.client()
        self.data_table: CollectionReference = self.db.collection(self.config.data_table)
        self.audit_table: CollectionReference = self.db.collection(self.config.audit_table)
        self.auth_table: CollectionReference = self.db.collection(self.config.auth_table)
        self.group_table: CollectionReference = self.db.collection(self.config.group_table)

    def get_auth(self, user_id: str) -> Optional[User]:
        # Try to find user in the auth table
        result = self.auth_table.document(user_id).get()

        if not result.exists:
            return None

        # Create user object
        return User.load(result.to_dict())

    def get_entitlements(self, entitlement_ids: List[str]) -> List[User]:
        entitlements = []
        for result in Query(self.auth_table).where('id', 'in', entitlement_ids).stream():
            record = result.to_dict()
            record['id'] = result.id
            entitlements.append(User.load(record))
        return entitlements

    def get_group(self, group_id: str) -> Optional[Group]:
        # Try to find user in the auth table
        result = self.group_table.document(group_id).get()

        if not result.exists:
            return None

        # Create user object
        return Group.load(result.to_dict())

    def store_item(self, table: str, item: dict) -> bool:
        try:
            self.db.collection(table).document().set(item)
        except Exception as e:
            print('Failed to store record: %s', e)
            return False

        return True

    def create(self, request: Request, data: dict, validation: dict = None,
               required_fields: Union[List, Tuple] = ()) -> dict:
        """
        Create an item in Dynamo

        :param Request request: Request object
        :param dict data: Item to create
        :param dict validation: Optional dictionary containing mappings of field name to callable. See the docstring
        in the _validate_fields method for more information.
        :param list required_fields: List of required fields
        :return: Created item
        :rtype: dict
        """
        user = self._prepare_create(request, data, validation, required_fields)

        # Make sure primary key is included
        document_id = data.get(self.config.primary_key)
        if not document_id:
            raise BadRequestException('Primary key %s is required' % self.config.primary_key)

        # Set the resource identifier
        resource = {self.config.primary_key: document_id}

        # TODO: Build condition to ensure the unique key does not exist

        # Save the item
        try:
            self.data_table.document(document_id).set(data)
            logger.info('[%(user)s] Successfully created item:\n%(item)s' % {
                'user': self.user_identifier(user),
                'item': data
            })
        except Exception as e:
            logger.error(
                '[%(user)s] Encountered error while attempting to create record %(error)s. Item:\n%(item)s' % {
                    'user': self.user_identifier(user),
                    'error': str(e),
                    'item': data
                }
            )
            raise

        if has_sentry:
            sentry_sdk.add_breadcrumb(category='data', message='Created item in Firestore', level='info')

        # Create audit log
        self.audit_log(action=self.AUDIT_ACTION_CREATE, resource=resource, request=request, user=user)

        return resource

    def update(self, request: Request, primary_key: dict, data: dict, validation: dict,
               audit_action=BaseAPI.AUDIT_ACTION_UPDATE) -> dict:
        """
        Update an item in Dynamo

        :param Request request: Request object
        :param dict primary_key: Dictionary formatted as {"partition_key": "value_of_row_to_update"}
        :param dict data: Fields to update, formatted as {"key": "value"}
        :param dict validation: Optional dictionary containing mappings of field name to callable. See the
        docstring in the _validate_fields method for more information.
        :param str audit_action: Action to use in the audit log. This defaults to UPDATE, and is provided as a
        convenience to the user in case customizing the phrasing is desired. This cannot be one of the reserved
        built-in actions: CREATE, DELETE, GET, LIST, SEARCH.
        :return: Updated item
        :rtype: dict
        """
        # Get user
        user = self.initialize_request(request)

        # Validate audit action
        audit_action = audit_action.upper()
        if audit_action in (self.AUDIT_ACTION_CREATE, self.AUDIT_ACTION_DELETE,
                            self.AUDIT_ACTION_GET, self.AUDIT_ACTION_LIST, self.AUDIT_ACTION_SEARCH):
            raise Exception('%s is a reserved built-in audit action' % audit_action)

        # Deny updates to primary key
        if self.config.primary_key in data:
            raise BadRequestException('Primary key %s cannot be updated' % self.config.primary_key)

        # Add in the user's permissions
        filtering = GCPFiltering(self.data_table)
        filtering.filter(user, action=filtering.FILTER_ACTION_READ)

        # Get the existing item
        existing_item = filtering.query.stream()
        if len(existing_item) == 0:
            logger.info('[%(user)s] Primary key "%(partition_key)s" does not exist or user does '
                        'not have permission to access it' % {
                            'user': self.user_identifier(user),
                            'partition_key': primary_key
                        })
            raise NotFoundException('Item does not exist or you do not have permission to access it')
        elif len(existing_item) > 1:
            # Should not have found more than one item
            raise BadRequestException('Multiple items found')

        # Found the item
        existing_item = existing_item[0]

        # Validate update
        self.validate_update(user, data, existing_item)

        # Perform field validation
        if validation:
            self.validate_fields(
                validation=validation,
                required_fields=(),
                item=data,
                existing_item=existing_item
            )
            if has_sentry:
                sentry_sdk.add_breadcrumb(category='validate', message='Validated input fields', level='info')

        # Perform the update item call
        try:
            document_id = primary_key[self.config.primary_key]
            self.data_table.document(document_id).update(data)
            response = self.data_table.document(document_id).get()
            logger.info('[%(user)s] Successfully updated record "%(partition_key)s" with values:\n%(item)s' % {
                'user': self.user_identifier(user),
                'partition_key': primary_key,
                'item': data
            })
        except Exception as e:
            logger.error(
                '[%(user)s] Encountered error while attempting to update record '
                '"%(partition_key)s": %(error)s. Item:\n%(item)s' % {
                    'user': self.user_identifier(user),
                    'partition_key': primary_key,
                    'error': str(e),
                    'item': data
                }
            )
            raise

        if has_sentry:
            sentry_sdk.add_breadcrumb(category='data', message='Updated item in Dynamo', level='info')

        # Create audit log
        self.audit_log(action=audit_action, resource=primary_key, changes=data, request=request, user=user)

        # Return updated record
        return self.post_process([response.to_dict()], user)[0]

    def get(self, request: Request, key: Any, value: Any) -> dict:
        """
        Get a specific item from the table

        :param Request request: Request object
        :param str key: Key to search on
        :param str value: Value to search for
        :return: Item
        :rtype: dict
        """
        # Get user
        user = self.initialize_request(request)

        # Filter the data according to the user's permissions
        filtering = GCPFiltering(self.data_table)
        conditions = filtering.filter(user)
        filtering.And(
            conditions,
            filtering.equals(key, value)
        )

        # Search for the item
        data = filtering.query.stream()
        if has_sentry:
            sentry_sdk.add_breadcrumb(
                category='query', message='%s = %s' % (key, value), level='info', table=self.config.data_table
            )

        # Filter the response
        output = self.post_process(data, user)

        # There should only be a single item returned
        if len(output) == 0:
            raise NotFoundException('Item not found')
        elif len(output) > 1:
            # We should not be returning more than a single item
            raise BadRequestException('Multiple items returned')
        else:
            # Item was found, return the single item
            self.audit_log(action=self.AUDIT_ACTION_GET, request=request, user=user, resource={key: value})
            return output[0]

    def list(self, request: Request) -> List[dict]:
        """
        List all values in a table

        :param Request request: Request object
        :return: Data from the table
        :rtype: dict
        """
        user, params = self._prepare_list(request)

        args = {}
        filtering = GCPFiltering(self.data_table)
        conditions = filtering.filter(user, params)
        if conditions:
            args.update({'FilterExpression': conditions})

        data = []
        for record in filtering.query.stream():
            record_dict = record.to_dict()
            record_dict[self.config.primary_key] = record.id
            data.append(record_dict)
        data = self.post_process(data, user)

        self.audit_log(self.AUDIT_ACTION_LIST, request, user)

        return data

    def list_unique_values(self, request: Request, key: str,
                           unique_func: Callable[[List[Dict], str], List[str]] = BaseAPI.unique_func) -> List[str]:
        user, params = self._prepare_list(request, False)

        # Build filters
        filtering = GCPFiltering(self.data_table)
        filtering.filter(user, params)

        # TODO: Add key existence filter

        # Download the data
        try:
            data = filtering.query.stream()
        except Exception as e:
            logger.error(
                '[%(user)s] Encountered error while attempting to list records: %(error)s' % {
                    'user': self.user_identifier(user),
                    'error': str(e)
                }
            )
            raise

        # Post process the data
        data = self.post_process(data, user)

        # Make sure a unique, sorted list is returned
        output = unique_func(data, key)

        # Create audit log
        self.audit_log(self.AUDIT_ACTION_LIST, request, user)

        return output

    def list_audit_logs(self, request: Request, param_overrides: dict = None) -> List[dict]:
        """
        List all audit logs

        :param Request request: Request object
        :param dict param_overrides: Parameter overrides: Each parameter should be formatted as:

            {
                "field_name": "search_value"
            }

        :return: List of audit logs, sorted with most recent entry first
        :rtype: list of dict
        """
        if not self.config.audit_table:
            raise NotFoundException('Audit logs are not enabled')

        if param_overrides is None:
            param_overrides = {}

        # Merge parameter overrides into query params
        request.query_params.update(param_overrides)

        # Get user
        user = self.initialize_request(request)

        # Build filters
        filtering = GCPFiltering(self.audit_table)
        filtering.filter(None, request.query_params)

        # Download data
        try:
            data = filtering.query.stream()
        except Exception as e:
            logger.error(
                '[%(user)s] Encountered error while attempting to list audit logs: %(error)s' % {
                    'user': self.user_identifier(user),
                    'error': str(e)
                }
            )
            raise

        # Add sentry breadcrumb
        if has_sentry:
            sentry_sdk.add_breadcrumb(
                category='query',
                message='Scanned table',
                level='info',
                table=self.config.audit_table
            )

        # Sort the data
        data = sorted(data, key=lambda item: item['time'], reverse=True)

        return data

    def search(self, request: Request, key: str, values: List[str]) -> List[dict]:
        """
        Perform a multi-value search of a field in the table. The search endpoint should be configured in API Gateway:

            POST /search/{search_key}

        The value of {search_key} should be passed in as `key` and the contents of the POST request body should be
        passed in as the `values` list.

        :param Request request: Request object
        :param str key: Field to search
        :param list of str values: Search values
        :return: Search results
        :rtype: dict
        """
        # Get user
        user = self.initialize_request(request)

        # Build multi-value filter expressions
        filtering = GCPFiltering(self.data_table)
        filters = filtering.multi_filter(user, key, values)
        if has_sentry:
            sentry_sdk.add_breadcrumb(category='data', message='Built multi-value filter', level='info')

        # Perform each generated filter expression and then combine the results together
        output = []
        for f in filters:
            for record in f.stream():
                # Add to output
                record_dict = record.to_dict()
                record_dict[self.config.primary_key] = record.id
                output.append(record_dict)

        # Create audit log
        self.audit_log(action=self.AUDIT_ACTION_SEARCH, request=request, user=user)

        # Return the filtered response
        return self.post_process(output, user)

    def delete(self, request: Request, primary_key: dict) -> dict:
        """
        Delete an item

        :param Request request: Request object
        :param dict primary_key: Dictionary formatted as {"primary_key": "value_of_row_to_delete"}
        :return: Success
        :rtype: dict
        """
        # Get user
        user = self.initialize_request(request)

        # TODO: Default conditional expression to make sure the item actually exists

        # Add in the user's permissions
        filtering = GCPFiltering(self.data_table)
        filtering.filter(user, action=filtering.FILTER_ACTION_DELETE)

        # Perform the deletion
        try:
            self.data_table.document(primary_key[self.config.primary_key]).delete()
            logger.info('[%(user)s] Successfully deleted record "%(primary_key)s"' % {
                'user': self.user_identifier(user),
                'primary_key': primary_key,
            })
        except Exception as e:
            logger.error(
                '[%(user)s] Encountered error while attempting to delete record "%(primary_key)s": %(error)s' % {
                    'user': self.user_identifier(user),
                    'primary_key': primary_key,
                    'error': str(e)
                }
            )
            raise

        if has_sentry:
            sentry_sdk.add_breadcrumb(category='data', message='Deleted item', level='info')

        # Create audit log
        self.audit_log(
            action=self.AUDIT_ACTION_DELETE,
            request=request,
            user=user,
            resource=primary_key,
        )

        # Return updated record
        return {'deleted': True}



from typing import List, Dict, Union, Any

from scoutr.exceptions import InvalidUserException
from scoutr.models import Model


class FilterField(Model):
    field: str
    operator: str = 'eq'
    value: Union[List, str]

    @classmethod
    def load(cls, data: Dict[str, Any]):
        filter_field = super(FilterField, cls).load(data)

        # Perform type check
        if not isinstance(data.get('field', None), str):
            raise InvalidUserException(f'Invalid type for field on user filter fields')
        if not isinstance(data.get('value', None), (str, list)):
            raise InvalidUserException(f'Invalid type for value on user filter fields')
        if isinstance(data.get('value', None), list):
            for item in data:
                if not isinstance(item, str):
                    raise InvalidUserException(f'User filter fields value must be a list of string values')
        if not isinstance(data.get('operator', ''), str):
            raise InvalidUserException(f'Invalid type for operator on user filter fields')

        return filter_field


class PermittedEndpoints(Model):
    endpoint: str
    method: str

    @classmethod
    def load(cls, data: Dict[str, str]):
        permitted_endpoint = super(PermittedEndpoints, cls).load(data)

        # Perform type check
        if not isinstance(data['endpoint'], str):
            raise InvalidUserException('Invalid type for endpoint on user permitted endpoints')
        if not isinstance(data['method'], str):
            raise InvalidUserException('Invalid type for method on user permitted endpoints')

        return permitted_endpoint


class Permissions(Model):
    permitted_endpoints: List[PermittedEndpoints] = []
    read_filters: List[FilterField] = []
    create_filters: List[FilterField] = []
    update_filters: List[FilterField] = []
    delete_filters: List[FilterField] = []
    exclude_fields: List[str] = []
    update_fields_permitted: List[str] = []
    update_fields_restricted: List[str] = []

    def __init__(self, permitted_endpoints: List[Dict[str, str]] = None, read_filters: List[dict] = None,
                 create_filters: List[dict] = None, update_filters: List[dict] = None,
                 delete_filters: List[dict] = None, exclude_fields: List[str] = None,
                 update_fields_permitted: List[str] = None, update_fields_restricted: List[str] = None, **kwargs):
        super(Permissions, self).__init__(**kwargs)

        if not read_filters:
            read_filters = []
        if not create_filters:
            create_filters = []
        if not update_filters:
            update_filters = []
        if not delete_filters:
            delete_filters = []
        if not permitted_endpoints:
            permitted_endpoints = []
        if not exclude_fields:
            exclude_fields = []
        if not update_fields_permitted:
            update_fields_permitted = []
        if not update_fields_restricted:
            update_fields_restricted = []

        for item in read_filters:
            self.read_filters.append(FilterField.load(item))

        for item in create_filters:
            self.create_filters.append(FilterField.load(item))

        for item in update_filters:
            self.update_filters.append(FilterField.load(item))

        for item in delete_filters:
            self.delete_filters.append(FilterField.load(item))

        for item in permitted_endpoints:
            self.permitted_endpoints.append(PermittedEndpoints.load(item))

        self.exclude_fields = exclude_fields
        self.update_fields_permitted = update_fields_permitted
        self.update_fields_restricted = update_fields_restricted


class User(Permissions):
    id: str
    name: str
    username: str
    email: str
    groups: List[str]


class Group(Permissions):
    id: str



from setuptools import setup, find_packages

with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name='scoutr',
    description='Generic full access control API for talking to a NoSQL database',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords='dynamo firestore mongo nosql api rbac',
    url='https://github.com/GESkunkworks/scoutr',
    use_scm_version=True,
    author='Michael Palmer',
    packages=find_packages(),
    setup_requires=[
        'setuptools_scm>=3.3.3,<4.2.0'
    ],
    install_requires=[],
    extras_require={
        'dev': [
            'mypy'
        ],
        'flask': [
            'Flask-API>=1.1,<3.0',
            'Flask~=1.1.1',
            'simplejson>=3.16,<3.18'
        ],
        'dynamo': [
            'boto3~=1.14.37',
            'botocore~=1.17.37',
            'boto3-stubs[dynamodb]'
        ],
        'firestore': [
            'google~=3.0.0',
            'firebase-admin~=4.4.0',
        ],
        'mongo': [
            'pymongo~=3.11.0'
        ],
        'docs': [
            'mkdocs',
            'mkdocs-material'
        ]
    },
    classifiers=[
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Development Status :: 5 - Production/Stable'
    ]
)



from typing import Tuple, Union


class Config:
    data_table: str
    auth_table: str
    group_table: str
    audit_table: str = ''
    primary_key: str = 'id'
    log_retention_days: int = 30
    oidc_username_header: str = ''
    oidc_name_header: Union[str, Tuple[str, str]] = ''
    oidc_email_header: str = ''
    oidc_group_header: str = ''

    def __init__(self, data_table: str, auth_table: str, group_table: str, audit_table: str = '',
                 primary_key: str = 'id', log_retention_days: int = 30, oidc_username_header: str = '',
                 oidc_name_header: Union[str, Tuple[str, str]] = '', oidc_email_header: str = '',
                 oidc_group_header: str = ''):
        self.data_table = data_table
        self.auth_table = auth_table
        self.audit_table = audit_table
        self.group_table = group_table
        self.primary_key = primary_key
        self.log_retention_days = log_retention_days
        self.oidc_username_header = oidc_username_header
        self.oidc_name_header = oidc_name_header
        self.oidc_email_header = oidc_email_header
        self.oidc_group_header = oidc_group_header


class MongoConfig(Config):
    connection_string: str
    database: str

    def __init__(self, connection_string: str, database: str, data_table: str, auth_table: str, group_table: str,
                 audit_table: str = '', primary_key: str = 'id', log_retention_days: int = 30,
                 oidc_username_header: str = '', oidc_name_header: Union[str, Tuple[str, str]] = '',
                 oidc_email_header: str = '', oidc_group_header: str = ''):
        super(MongoConfig, self).__init__(
            data_table=data_table,
            auth_table=auth_table,
            group_table=group_table,
            audit_table=audit_table,
            primary_key=primary_key,
            log_retention_days=log_retention_days,
            oidc_username_header=oidc_username_header,
            oidc_name_header=oidc_name_header,
            oidc_email_header=oidc_email_header,
            oidc_group_header=oidc_group_header
        )
        self.connection_string = connection_string
        self.database = database



from functools import wraps
from typing import List, Dict
from urllib.parse import parse_qsl

from flask_api.exceptions import NotFound, PermissionDenied, APIException, ParseError
from flask_api.request import APIRequest

from scoutr.exceptions import HttpException, UnauthorizedException, NotFoundException, BadRequestException
from scoutr.helpers.flask.oidc import get_user_from_oidc
from scoutr.models.request import RequestUser
from scoutr.providers.base.api import BaseAPI

try:
    import sentry_sdk
    has_sentry = True
except ImportError:
    has_sentry = False


def flaskapi_exception_wrapper(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except HttpException as e:
            if isinstance(e, UnauthorizedException):
                raise PermissionDenied(e.args)
            elif isinstance(e, NotFoundException):
                raise NotFound(e.args)
            elif isinstance(e, BadRequestException):
                raise ParseError(e.args)
            else:
                if has_sentry:
                    sentry_sdk.capture_exception(e)
                raise APIException(e.args)
        except Exception as e:
            if has_sentry:
                sentry_sdk.capture_exception(e)
            raise APIException('%s: %s' % (e.__class__.__name__, str(e)))
    return wrapper


def parse_query_params(query_string: bytes) -> Dict[str, List[str]]:
    params: Dict[str, List[str]] = {}

    for key, value in parse_qsl(query_string):
        key_str = key.decode('utf-8')
        value_str = value.decode('utf-8')
        params.setdefault(key_str, [])
        params[key_str].append(value_str)

    return params


def init_flask_user(api: BaseAPI, request: APIRequest) -> RequestUser:
    user = get_user_from_oidc(api, request)
    if has_sentry:
        import sentry_sdk
        with sentry_sdk.configure_scope() as scope:
            scope.set_user({
                'id': user.id,
                'ip_address': request.remote_addr,
                'username': user.data.username,
                'name': user.data.name,
                'email': user.data.email,
                'groups': user.data.groups
            })
    return user
